from datetime import datetime
from typing import List

import numpy as np
from pandas import DataFrame, NA, MultiIndex, concat

from .._tools import ohlc


def merge_dataframes(dfs: List[DataFrame]) -> DataFrame:
    df = dfs.pop()
    df = df.join(dfs, how="outer")  # noqa
    df = df.convert_dtypes()
    df.index.name = "Timestamp"

    df.ohlc = ohlc.__get__(df, None)
    return df


def create_df(
    data: list, timestamps: list, fields: list, stream_name: str
) -> DataFrame:
    numpy_array = np.array(data)
    timestamp_array = np.array(timestamps)

    if np.size(numpy_array):
        dataframe = DataFrame(numpy_array, columns=fields, index=timestamp_array)
    else:
        dataframe = DataFrame()

    dataframe.sort_index(inplace=True)
    dataframe.columns = MultiIndex.from_product([[stream_name], dataframe.columns])
    return dataframe


def retrieve_data_for_df(stream_data: List[dict], repeat: bool = False) -> tuple:
    timestamps = []
    data = []
    fields = set()

    fields.update(*(item["Fields"] for item in stream_data))
    fields = list(fields)

    for idx, record in enumerate(stream_data):
        if repeat and idx == 0:
            timestamps.append(datetime.now())
        else:
            timestamps.append(record["Timestamp"])

        rics_data = [record["Fields"].get(field) for field in fields]
        data.append(rics_data)

    return timestamps, data, fields


def replace_values_by_nan(_data):
    for rics_data in _data:
        for idx, value in enumerate(rics_data):
            try:
                if value is not None:
                    float(value)
            except ValueError:
                rics_data[idx] = None


def create_df_based_on_ticks(df):
    first_update = df.index[0]
    last_update = df.index[-1]
    time_delta = last_update - first_update

    if hasattr(time_delta, "days") and time_delta.days != 0:
        days = time_delta.days + 1
        days = str(days) + "D"
        df = df.ohlc(days, origin="end", call_from_recorder=True)
    elif hasattr(time_delta, "hours"):
        hours = time_delta.hour + 1
        hours = str(hours) + "H"
        df = df.ohlc(hours, origin="end", call_from_recorder=True)
    elif hasattr(time_delta, "minutes"):
        minutes = time_delta.minutes + 1
        minutes = str(minutes) + "min"
        df = df.ohlc(minutes, origin="end", call_from_recorder=True)
    else:
        seconds = time_delta.seconds + 1
        seconds = str(seconds) + "s"
        df = df.ohlc(seconds, origin="end", call_from_recorder=True)

    return df


class OHLCBuilder:
    def __init__(self, frequency: str, universe: list, fields: list) -> None:
        self._frequency = frequency
        self.dataframes = []
        self.ohlc_df = None
        self._universe = universe
        self._fields = fields

        self._last_recorded_ohlc_updates = None

    def create_ohlc_df(self, df: DataFrame) -> DataFrame:
        return df.ohlc(self._frequency, origin="end", call_from_recorder=True)

    def save_ohlc_data(self, df: DataFrame):
        df.fillna(NA, inplace=True)

        if self.ohlc_df is None:
            self.ohlc_df = df

        else:
            self.ohlc_df = concat([self.ohlc_df, df])

    def build(self, updates_by_stream_name: dict):
        dfs = []
        df = DataFrame()
        self._count_of_updates = 0

        for universe, stream_data in updates_by_stream_name.items():
            _timestamps, _data, _fields = retrieve_data_for_df(stream_data, True)
            self._count_of_updates += len(_data)
            replace_values_by_nan(_data)
            updates_by_stream_name[universe] = []
            dataframe = create_df(_data, _timestamps, _fields, universe)
            dfs.append(dataframe)

        if dfs:
            df = merge_dataframes(dfs)
            self.dataframes.append(df)

        if not df.empty:
            df = self.create_ohlc_df(df)
            df["ticks count"] = self._count_of_updates

        if df.empty:
            empty_row = {}

            if isinstance(self.ohlc_df, DataFrame) and not self.ohlc_df.empty:
                columns = [col for col in self.ohlc_df]
            else:
                columns = self._create_ohlc_columns()

            for column in columns:
                empty_row[column] = np.nan

            _timestamps = np.array([datetime.now()])
            df = DataFrame(empty_row, index=_timestamps)

        self._last_recorded_ohlc_updates = df
        self.save_ohlc_data(df)

    def _create_ohlc_columns(self):
        columns = []
        ohlc_list = ["open", "high", "low", "close"]

        for universe in self._universe:
            for field in self._fields:
                for item in ohlc_list:
                    columns.append((universe, field, item))

        columns.append(("ticks count", "", ""))
        return columns

    def dispose(self):
        self.dataframes = []
        self.ohlc_df = None


class Ticks_OHLCBuilder(OHLCBuilder):
    def create_ohlc_df(self, df: DataFrame) -> DataFrame:
        return create_df_based_on_ticks(df)



import pytest

from refinitiv.data._tools import inspect_parameters_without_self
from refinitiv.data.delivery.cfs import packages
from tests.unit.conftest import StubSession
from tests.unit.delivery.cfs.conftest import PACKAGES_RESPONSES
import refinitiv.data as rd


@pytest.mark.parametrize(
    "value",
    [
        "get_data",
        "get_data_async",
    ],
)
def test_attribute(value):
    # given

    # when
    definition = packages.Definition("")

    # then
    assert hasattr(definition, value)


def test_fist_argument():
    # when
    args_class = list(inspect_parameters_without_self(packages.Definition))
    test_first_arg = args_class[0]

    # then
    assert test_first_arg == "package_name"


@pytest.mark.parametrize(
    "is_default_session",
    [True, False],
    ids=[
        "use default session",
        "without default session",
    ],
)
def test_iteration_by_packages_without_error(is_default_session):
    # given
    session = StubSession(is_open=True, response=PACKAGES_RESPONSES)
    definition = packages.Definition("Bulk-ESG-Global-Scores-Wealth-Standard-v1")

    if is_default_session:
        rd.session.set_default(session)
        response = definition.get_data()

    else:
        response = definition.get_data(session=session)

    # when
    try:
        for package in response.data.packages:
            for fileset in package:
                value = fileset
    except Exception as e:
        rd.session.set_default(None)
        assert False, str(e)

    # then
    else:
        rd.session.set_default(None)
        assert True



# This test is commented due to temporary removing of
# function from publick API
# https://jira.refinitiv.com/browse/EAPI-2233
# Uncomment it if function would be returned back

# """
# https://jira.refinitiv.com/browse/EAPI-2223
# """
#
# import inspect
#
# import pytest
#
# import refinitiv.data as rd
#
# unneeded_params = ["on_response", "session", "closure"]
#
#
# @pytest.mark.parametrize(
#     "testing_function",
#     [
#         getattr(rd.function, func_name)
#         for func_name in dir(rd.function)
#         if not func_name.startswith("_")
#     ],
# )
# def test_function(testing_function):
#     argspec = inspect.getfullargspec(testing_function)
#
#     testing_unneeded_params = filter(lambda arg: arg in unneeded_params, argspec.args)
#
#     assert list(testing_unneeded_params) == [], testing_function.__name__



# coding: utf8

__all__ = ["EndOfMonthConvention"]

from enum import Enum, unique


@unique
class EndOfMonthConvention(Enum):
    """
    End of month convention.

    The possible values are:
        - Last,
        - Same,
        - Last28,
        - Same28
    """

    LAST = "Last"
    LAST28 = "Last28"
    SAME = "Same"
    SAME28 = "Same28"



from unittest.mock import Mock

from refinitiv.data.content.ipa.surfaces import fx


def test_surfaces_fx_definition_repr():
    # given
    mock_surface_layout = Mock()
    mock_surface_parameters = Mock()
    definition = fx.Definition(
        surface_layout=mock_surface_layout,
        surface_parameters=mock_surface_parameters,
        underlying_definition={},
    )
    expected_value = (
        f"<refinitiv.data.content.ipa.surfaces.fx.Definition "
        f"object at {hex(id(definition))}>"
    )

    # when
    testing_value = repr(definition)

    # then
    assert testing_value == expected_value


def test_surfaces_fx_definition_output_repr():
    # given
    definition = fx.SurfaceLayout()
    expected_value = (
        f"<refinitiv.data.content.ipa.surfaces.fx.SurfaceLayout "
        f"object at {hex(id(definition))}>"
    )

    # when
    testing_value = repr(definition)

    # then
    assert testing_value == expected_value



from unittest import mock

import pytest

import refinitiv.data as rd
from refinitiv.data._config_functions import _load_config_and_set_default
from refinitiv.data._configure import _RDPConfig


@mock.patch("os.path.exists")
def test_load_config(mock_file_exist):
    path = "path/to/file.json"
    result = rd.load_config(path)
    assert result.__class__.__name__ == _RDPConfig.__name__


@mock.patch("os.path.exists")
def test_load_config_and_set_default(mock_file_exist):
    path = "path/to/file.json"
    result = _load_config_and_set_default(path)
    assert result.__class__.__name__ == _RDPConfig.__name__


def test_load_config_error():
    with pytest.raises(FileNotFoundError) as error:
        rd.load_config("path/to/file.json")
    assert "Can't find file: path/to/file.json." in error.value.args[0]


def test_get_config():
    config = rd.get_config()
    assert config.__class__.__name__ == _RDPConfig.__name__


def test_set_param():
    config = rd.get_config()
    config.set_param("custom.param", "custom-value", auto_create=True)
    assert config.get_param("custom.param") == "custom-value"


def test_set_param_error():
    config = rd.get_config()
    with pytest.raises(
        ValueError,
        match="'invalid.param' is not defined in config. "
        "To create new parameter please use 'auto_create' property.",
    ):
        config.set_param("invalid.param", "custom-value")



from refinitiv.data.content import historical_pricing
from refinitiv.data.content._intervals import Intervals
from refinitiv.data.content.historical_pricing import MarketSession
from tests.unit.conftest import StubSession
from tests.unit.content.historical_pricing.data_for_tests import (
    DF_WITHOUT_NAN_RESPONSES,
    DF_REQUIREMENT_TWO_INSTS_ONE_FIELD,
)


def test_df_without_nan():
    # given
    session = StubSession(is_open=True, response=DF_WITHOUT_NAN_RESPONSES)
    expected_str = (
        "                       AMD.O                                                  IMNM.O                                          \n"
        "                    OPEN_PRC   HIGH_1    LOW_1 TRDPRC_1 NUM_MOVES ACVOL_UNS OPEN_PRC HIGH_1 LOW_1 TRDPRC_1 NUM_MOVES ACVOL_UNS\n"
        "Timestamp                                                                                                                     \n"
        "2022-05-25 19:54:00     <NA>     <NA>     <NA>     <NA>      <NA>      <NA>        3      3     3        3         1       104\n"
        "2022-05-25 19:56:00     <NA>     <NA>     <NA>     <NA>      <NA>      <NA>        3      3     3        3         2       224\n"
        "2022-05-25 19:57:00    92.55   92.575    92.44    92.54      1186    306126        3      3     3        3         1       100\n"
        "2022-05-25 19:58:00    92.53  92.5714  92.4722  92.5575      1451    339267     <NA>   <NA>  <NA>     <NA>      <NA>       100\n"
        "2022-05-25 19:59:00    92.55    92.71   92.484  92.7099      2483    619187     <NA>   <NA>  <NA>     <NA>      <NA>       231\n"
        "2022-05-25 20:00:00     92.7     92.7    92.65    92.65         2   1750467     <NA>   <NA>  <NA>     <NA>      <NA>      <NA>\n"
        "2022-05-25 20:01:00     <NA>     <NA>     <NA>     <NA>      <NA>     56358     <NA>   <NA>  <NA>     <NA>      <NA>      <NA>"
    )
    definition = historical_pricing.summaries.Definition(
        universe=["AMD.O", "IMNM.O"],
        interval=Intervals.ONE_MINUTE,
        start="2022-05-25T00:00:00",
        end="2022-05-25T23:59:10",
        count=5,
        fields=["OPEN_PRC", "HIGH_1", "LOW_1", "TRDPRC_1", "NUM_MOVES", "ACVOL_UNS"],
        sessions=[
            MarketSession.NORMAL,
        ],
    )
    # when
    response = definition.get_data(session=session)
    df = response.data.df

    # then
    assert df.to_string() == expected_str


def test_df_requirement_two_insts_one_field():
    # given
    session = StubSession(is_open=True, response=DF_REQUIREMENT_TWO_INSTS_ONE_FIELD)
    expected_str = (
        "BID           GBP=  G234BP=\n"
        "Date                       \n"
        "2022-06-09   1.249     <NA>\n"
        "2022-06-10  1.2314     <NA>\n"
        "2022-06-13  1.2134     <NA>\n"
        "2022-06-14  1.1993     <NA>\n"
        "2022-06-15  1.2178     <NA>\n"
        "2022-06-16  1.2351     <NA>\n"
        "2022-06-17  1.2224     <NA>\n"
        "2022-06-20   1.225     <NA>\n"
        "2022-06-21  1.2272     <NA>\n"
        "2022-06-22  1.2266     <NA>"
    )
    definition = historical_pricing.summaries.Definition(
        universe=["GBP=", "G234BP="], interval=Intervals.DAILY, count=10, fields=["BID"]
    )
    # when
    response = definition.get_data(session=session)
    df = response.data.df

    # then
    assert df.to_string() == expected_str


def test_fields_is_not_change():
    expected_fields = ["BID"]
    testing_fields = ["BID"]
    session = StubSession(is_open=True, response=DF_REQUIREMENT_TWO_INSTS_ONE_FIELD)
    definition = historical_pricing.summaries.Definition(
        universe=["GBP=", "G234BP="],
        interval=Intervals.DAILY,
        count=10,
        fields=testing_fields,
    )
    definition.get_data(session=session)
    assert testing_fields == expected_fields



import abc
from typing import TYPE_CHECKING, Optional, List

from ._data_classes import (
    HeadlineRDP,
    Story,
    StoryUDF,
    HeadlineUDF,
)
from ._sort_order import SortOrder
from ._tools import get_headlines
from .._content_data import Data
from .._content_data_provider import ContentDataProvider
from .._content_response_factory import ContentResponseFactory
from .._join_responses import join_responses
from ..._core.session import DesktopSession
from ..._tools import to_datetime, make_enum_arg_parser, extend_params
from ...delivery._data._data_provider import (
    RequestFactory,
    Response,
    ContentValidator,
    ValidatorContainer,
    EndpointData,
)
from ...delivery._data._endpoint_data import RequestMethod
from ...delivery._data._response_factory import ResponseFactory

if TYPE_CHECKING:
    from ...delivery._data._data_provider import ParsedData
    from ..._types import OptInt

MAX_LIMIT = 100


# --------------------------------------------------------------------------------------
#   Content data
# --------------------------------------------------------------------------------------


# NewsStory
class NewsStoryData(EndpointData):
    _story: Optional[Story] = None

    @abc.abstractmethod
    def _build_story(self, raw: dict) -> Story:
        # override this
        pass

    @property
    def story(self) -> Story:
        if self._story is None:
            self._story = self._build_story(self.raw)
        return self._story


class NewsStoryRDPData(NewsStoryData):
    def _build_story(self, raw: dict) -> Story:
        return Story.from_dict(raw)


class NewsStoryUDFData(NewsStoryData):
    def _build_story(self, raw: dict) -> StoryUDF:
        return StoryUDF.from_dict(raw)


# NewsHeadlines


class NewsHeadlinesData(Data):
    _headlines: Optional[List[HeadlineRDP]] = None
    _limit: "OptInt" = None

    @abc.abstractmethod
    def _build_headlines(self, raw: dict, limit: int) -> List[HeadlineRDP]:
        # override this
        pass

    @property
    def headlines(self) -> List[HeadlineRDP]:
        if self._headlines is None:
            self._headlines = self._build_headlines(self.raw, self._limit)

        return self._headlines


class NewsHeadlinesRDPData(NewsHeadlinesData):
    def _build_headlines(self, raw: dict, limit: int) -> List[HeadlineRDP]:
        return get_headlines(raw, HeadlineRDP.from_dict, limit)


class NewsHeadlinesUDFData(NewsHeadlinesData):
    def _build_headlines(self, raw: dict, limit: int) -> List[HeadlineUDF]:
        return get_headlines(raw, HeadlineUDF.from_dict, limit)


# --------------------------------------------------------------------------------------
#   Response object
# --------------------------------------------------------------------------------------


class NewsStoryResponse(Response):
    def __str__(self):
        if self.data.raw:
            return (
                self.data.raw.get("newsItem", {})
                .get("contentSet", {})
                .get("inlineData", [{}])[0]
                .get("$")
                or self.data.story.content.text
            )
        else:
            return f"{self.errors}"


# --------------------------------------------------------------------------------------
#   Content data validator
# --------------------------------------------------------------------------------------


class NewsUDFContentValidator(ContentValidator):
    def __init__(self) -> None:
        super().__init__()
        self.validators.append(self.content_data_has_no_error)


# --------------------------------------------------------------------------------------
#   Response factory
# --------------------------------------------------------------------------------------


class NewsStoryResponseFactory(ResponseFactory):
    @staticmethod
    def _try_write_error(parsed_data: "ParsedData"):
        new_error_msg = "Error while calling the NEP backend: Story not found"
        error_code = parsed_data.first_error_code
        error_msg = parsed_data.first_error_message

        if error_code == 400 or error_code == 404 and new_error_msg != error_msg:
            parsed_data.error_codes = 404
            parsed_data.error_messages = new_error_msg

    def create_fail(self, parsed_data: "ParsedData", **kwargs):
        self._try_write_error(parsed_data)
        return super().create_fail(parsed_data, **kwargs)


# --------------------------------------------------------------------------------------
#   Request factory
# --------------------------------------------------------------------------------------

# NewsStory


class NewsStoryUDFRequestFactory(RequestFactory):
    def get_body_parameters(self, session, *args, **kwargs):
        entity = {
            "E": "News_Story",
        }
        w = dict()

        story_id = kwargs.get("story_id")
        w["storyId"] = story_id

        app_key = session.app_key
        w["productName"] = app_key

        entity["W"] = w
        body_parameters = {"Entity": entity}
        return body_parameters

    def get_url(self, session, *args, **kwargs):
        url = session._get_rdp_url_root()
        if isinstance(session, DesktopSession):
            url = session._get_udf_url()
        return url

    def update_url(self, url_root, url, path_parameters, query_parameters):
        return url

    def get_request_method(self, **kwargs) -> RequestMethod:
        return RequestMethod.POST


class NewsStoryRDPRequestFactory(RequestFactory):
    def get_path_parameters(self, session=None, *, story_id=None, **kwargs):
        return {"storyId": story_id}

    def get_header_parameters(self, session=None, **kwargs):
        return {"accept": "application/json"}

    def get_url(self, *args, **kwargs):
        return super().get_url(*args, **kwargs) + "/{storyId}"


# NewsHeadlines
class NewsHeadlinesUDFRequestFactory(RequestFactory):
    def extend_body_parameters(self, body_parameters, extended_params=None, **kwargs):
        if extended_params:
            body_parameters["Entity"]["W"].update(extended_params)
        return body_parameters

    def get_body_parameters(self, session, *args, **kwargs):
        entity = {
            "E": "News_Headlines",
        }
        w = dict()

        query = kwargs.get("query")
        w["query"] = query

        count = kwargs.get("count")
        if count is not None:
            w["number"] = str(count)

        payload = kwargs.get("payload")
        if payload is not None:
            w["payload"] = payload.replace("/headlines?payload=", "", 1)

        repository = kwargs.get("repository")
        if repository is not None:
            w["repository"] = repository

        app_key = session.app_key
        w["productName"] = app_key

        date_from = kwargs.get("date_from")
        if date_from is not None:
            w["dateFrom"] = to_datetime(date_from).isoformat()

        date_to = kwargs.get("date_to")
        if date_to is not None:
            w["dateTo"] = to_datetime(date_to).isoformat()

        entity["W"] = w
        body_parameters = {"Entity": entity}
        return body_parameters

    def get_url(self, session, *args, **kwargs):
        url = session._get_rdp_url_root()
        if isinstance(session, DesktopSession):
            url = session._get_udf_url()
        return url

    def update_url(self, url_root, url, path_parameters, query_parameters):
        return url

    def get_request_method(self, **kwargs) -> RequestMethod:
        return RequestMethod.POST


class NewsHeadlinesRDPRequestFactory(RequestFactory):
    def extend_query_parameters(self, query_parameters, extended_params=None):
        return extend_params(query_parameters, extended_params)

    def extend_body_parameters(self, body_parameters, extended_params=None, **kwargs):
        return body_parameters

    def get_query_parameters(self, *_, **kwargs):
        query_parameters = []

        query = kwargs.get("query")
        query_parameters.append(("query", query))

        count = kwargs.get("count")
        if count is not None:
            query_parameters.append(("limit", count))

        date_from = kwargs.get("date_from")
        if date_from is not None:
            date_from = to_datetime(date_from).isoformat()
            query_parameters.append(("dateFrom", date_from))

        date_to = kwargs.get("date_to")
        if date_to is not None:
            date_to = to_datetime(date_to).isoformat()
            query_parameters.append(("dateTo", date_to))

        sort_order = kwargs.get("sort_order")
        if sort_order is not None:
            sort_order = sort_order_news_arg_parser.get_str(sort_order)
            query_parameters.append(("sort", sort_order))

        # for pagination
        cursor = kwargs.get("cursor")
        if cursor is not None:
            query_parameters.append(("cursor", cursor))

        return query_parameters


# --------------------------------------------------------------------------------------
#   Data provider
# --------------------------------------------------------------------------------------

# NewsStory


class NewsUDFDataProvider(ContentDataProvider):
    @staticmethod
    def change_count(count: int, limit: int, kwargs: dict):
        number = abs(limit - count)
        if number < MAX_LIMIT:
            kwargs["count"] = number

    def get_data(self, *args, **kwargs):
        limit = kwargs.get("count")

        if limit is None:
            return super().get_data(*args, **kwargs)

        responses = []
        headlines = True
        count = 0

        if limit > MAX_LIMIT:
            kwargs["count"] = MAX_LIMIT

        while count < limit and headlines:
            response = super().get_data(*args, **kwargs)
            headlines = response.data.raw.get("headlines", [])
            count += len(headlines)
            kwargs["payload"] = response.data.raw.get("older")
            responses.append(response)
            self.change_count(count, limit, kwargs)

        response = join_responses(
            responses, data_class=self.response.data_class, limit=limit
        )
        response.data._limit = limit

        if not responses:
            response.is_success = True

        return response

    async def get_data_async(self, *args, **kwargs):
        limit = kwargs.get("count")

        if limit is None:
            return await super().get_data_async(*args, **kwargs)

        responses = []
        headlines = True
        count = 0

        if limit > MAX_LIMIT:
            kwargs["count"] = MAX_LIMIT

        while count < limit and headlines:
            response = await super().get_data_async(*args, **kwargs)
            headlines = response.data.raw.get("headlines", [])
            count += len(headlines)
            kwargs["payload"] = response.data.raw.get("older")
            responses.append(response)
            self.change_count(count, limit, kwargs)

        response = join_responses(
            responses, data_class=self.response.data_class, limit=limit
        )
        response.data._limit = limit

        if not responses:
            response.is_success = True

        return response


# NewsHeadlines


class NewsHeadlinesRDPDataProvider(ContentDataProvider):
    def get_data(self, *args, **kwargs):
        on_page_response = kwargs.get("on_page_response")
        limit = kwargs.get("count")
        responses = []
        cursor = True
        count = 0

        if limit > MAX_LIMIT:
            kwargs["count"] = MAX_LIMIT

        while count < limit and cursor:
            response = super().get_data(*args, **kwargs)
            responses.append(response)

            if on_page_response:
                on_page_response(self, response)

            meta = response.data.raw.get("meta", {})
            count += meta.get("count", 0)
            cursor = meta.get("next")
            kwargs = {
                "cursor": cursor,
                "__content_type__": kwargs.get("__content_type__"),
            }

        response = join_responses(
            responses, data_class=NewsHeadlinesRDPData, limit=limit
        )
        response.data._limit = limit

        if not responses:
            response.is_success = True

        return response

    async def get_data_async(self, *args, **kwargs):
        on_page_response = kwargs.get("on_page_response")
        limit = kwargs.get("count")
        responses = []
        cursor = True
        count = 0

        if limit > MAX_LIMIT:
            kwargs["count"] = MAX_LIMIT

        while count < limit and cursor:
            response = await super().get_data_async(*args, **kwargs)
            responses.append(response)

            if on_page_response:
                on_page_response(self, response)

            meta = response.data.raw.get("meta", {})
            count += meta.get("count", 0)
            cursor = meta.get("next")
            kwargs = {
                "cursor": cursor,
                "__content_type__": kwargs.get("__content_type__"),
            }

        response = join_responses(
            responses, data_class=NewsHeadlinesRDPData, limit=limit
        )
        response.data._limit = limit

        if not responses:
            response.is_success = True

        return response


sort_order_news_arg_parser = make_enum_arg_parser(SortOrder)

validator = ValidatorContainer(content_validator=NewsUDFContentValidator())

# NewsStory

news_story_data_provider_rdp = ContentDataProvider(
    response=NewsStoryResponseFactory(
        response_class=NewsStoryResponse,
        data_class=NewsStoryRDPData,
    ),
    request=NewsStoryRDPRequestFactory(),
)

news_story_data_provider_udf = NewsUDFDataProvider(
    response=NewsStoryResponseFactory(
        response_class=NewsStoryResponse,
        data_class=NewsStoryUDFData,
    ),
    request=NewsStoryUDFRequestFactory(),
    validator=validator,
)

# NewsHeadlines

news_headlines_data_provider_rdp = NewsHeadlinesRDPDataProvider(
    response=ContentResponseFactory(data_class=NewsHeadlinesRDPData),
    request=NewsHeadlinesRDPRequestFactory(),
)
news_headlines_data_provider_udf = NewsUDFDataProvider(
    response=ContentResponseFactory(data_class=NewsHeadlinesUDFData),
    request=NewsHeadlinesUDFRequestFactory(),
    validator=validator,
)

del validator



import allure
import pytest

import refinitiv.data as rd
from refinitiv.data.delivery import cfs
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.delivery.cfs.buckets.conftest import (
    check_buckets_in_response_created_after_date,
    check_buckets_in_response_modified_after_date,
    check_buckets_in_response_available_before_date,
    check_buckets_in_response_available_after_date,
    check_all_buckets_contain_attributes,
    check_buckets_in_two_responses_are_not_the_same,
)
from tests.integration.delivery.cfs.conftest import (
    check_amount_of_objects_in_response_not_bigger_than,
    check_no_objects_in_response,
)
from tests.integration.helpers import (
    check_response_status,
    check_non_empty_response_data,
    get_async_response_from_definition,
)


@allure.suite("Delivery object - CFS Buckets")
@allure.feature("Delivery object - CFS Buckets")
@allure.severity(allure.severity_level.CRITICAL)
class TestCfsBuckets:
    @allure.title("Get buckets with no params and default amount 25 - synchronously")
    @pytest.mark.caseid("36629292")
    @pytest.mark.smoke
    def test_get_buckets_with_no_params_synchronously(self, open_platform_session):
        response = cfs.buckets.Definition().get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_amount_of_objects_in_response_not_bigger_than(25, response)

    @allure.title("Get buckets with no params and default amount 25 - asynchronously")
    @pytest.mark.caseid("36629294")
    async def test_get_buckets_with_no_params_asynchronously(
        self, open_platform_session_async
    ):
        response = await get_async_response_from_definition(cfs.buckets.Definition())

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_amount_of_objects_in_response_not_bigger_than(25, response)

    @allure.title("Get buckets with closed session ")
    @pytest.mark.caseid("36629293")
    def test_get_buckets_with_closed_session(self):
        session = rd.session.desktop.Definition(app_key=" ").get_session()
        rd.session.set_default(session)
        buckets_definition = cfs.buckets.Definition()
        with pytest.raises(
            ValueError, match="Session is not opened. Can't send any request"
        ):
            buckets_definition.get_data()

    @allure.title("Get buckets with existing name param - partial match")
    @pytest.mark.caseid("C36629295")
    @pytest.mark.parametrize("name", ["cfs"])
    def test_get_buckets_with_existing_name_param(self, open_platform_session, name):
        response = cfs.buckets.Definition(name=name).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_amount_of_objects_in_response_not_bigger_than(25, response)

    @allure.title("Get buckets with non-existing name param")
    @pytest.mark.caseid("36629300")
    @pytest.mark.parametrize("name", ["Invalid_bucket_name"])
    def test_get_buckets_with_non_existing_name_param(
        self, open_platform_session, name
    ):
        response = cfs.buckets.Definition(name=name).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_no_objects_in_response(response)

    @allure.title("Get buckets with created_since param")
    @pytest.mark.caseid("36629552")
    @pytest.mark.parametrize("created_since", ["2020-10-23T00:00:00Z"])
    def test_get_buckets_with_created_since_param(
        self, open_platform_session, created_since
    ):
        response = cfs.buckets.Definition(created_since=created_since).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_buckets_in_response_created_after_date(created_since, response)

    @allure.title("Get buckets with modified_since param")
    @pytest.mark.caseid("36629552")
    @pytest.mark.parametrize("modified_since", ["2020-10-23T00:00:00Z"])
    def test_get_buckets_with_modified_since_param(
        self, open_platform_session, modified_since
    ):
        response = cfs.buckets.Definition(modified_since=modified_since).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_buckets_in_response_modified_after_date(modified_since, response)

    @allure.title("Get buckets with available_from param")
    @pytest.mark.caseid("36629591")
    @pytest.mark.parametrize("available_from", ["2020-10-23T00:00:00Z"])
    def test_get_buckets_with_available_from_param(
        self, open_platform_session, available_from
    ):
        response = cfs.buckets.Definition(available_from=available_from).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_buckets_in_response_available_before_date(available_from, response)

    @allure.title("Get buckets with available_to param")
    @pytest.mark.caseid("36629804")
    @pytest.mark.parametrize("available_to", ["2040-10-23T00:00:00Z"])
    def test_get_buckets_with_available_to_param(
        self, open_platform_session, available_to
    ):
        response = cfs.buckets.Definition(available_to=available_to).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_buckets_in_response_available_after_date(available_to, response)

    @pytest.mark.xfail(reason="https://jira.refinitiv.com/browse/EAPI-3100")
    @allure.title("Get buckets with list of attributes")
    @pytest.mark.parametrize(
        "attributes",
        [
            (["size", "depth", "purpose", "team"]),
            ("dayofweek"),
        ],
    )
    @pytest.mark.caseid("36629805")
    def test_get_buckets_with_list_of_attributes(
        self, open_platform_session, attributes
    ):
        response = cfs.buckets.Definition(attributes=attributes).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_all_buckets_contain_attributes(response, attributes)

    @allure.title("Get buckets with valid page_size param")
    @pytest.mark.caseid("36629806")
    @pytest.mark.parametrize("page_size", [26, 1000])
    def test_get_buckets_with_valid_page_size_param(
        self, open_platform_session, page_size
    ):
        response = cfs.buckets.Definition(page_size=page_size).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        buckets_count = len(response.data.raw["value"])
        assert (
            buckets_count <= page_size
        ), f"Buckets count received {buckets_count} more than expected: {page_size}"

    @allure.title("Get buckets with valid skip_token parameter")
    @pytest.mark.caseid("36629807")
    def test_get_buckets_with_valid_skip_token_param(self, open_platform_session):
        response1 = cfs.buckets.Definition().get_data()
        skip_token = response1.data.raw["skip_token"]
        response2 = cfs.buckets.Definition(skip_token=skip_token).get_data()
        check_response_status(response2, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response2)
        check_buckets_in_two_responses_are_not_the_same(response1, response2)



from refinitiv.data.content.ipa.curves import zc_curve_definitions


def zc_curve_definition_01():
    definition = zc_curve_definitions.Definition(
        source="Refinitiv",
        index_name="EURIBOR",
        valuation_date="2020-07-01",
    )
    return definition


def zc_curve_definition_02():
    return zc_curve_definitions.Definition()


def invalid_zc_curve_definition():
    definition = zc_curve_definitions.Definition(
        source="Invalid",
        index_name="EURIBOR",
        valuation_date="2020-07-01",
    )
    return definition


def check_sc_curve_definitions_response(response, expected_currency):
    actual_currency = response.data.raw["data"][0]["curveDefinitions"][0]["currency"]
    assert actual_currency == expected_currency, f"{actual_currency}"
    assert not response.data.df.isnull().values.all()


zc_curve_definition_universe = {
    "indexName": "EURIBOR",
    "source": "Refinitiv",
    "valuationDate": "2020-07-01",
}



from typing import TYPE_CHECKING

from ......._tools import try_copy_to_list
from .._mixin_request import MixinRequest
from ._curve_update_definition import CrossCurrencyCurveUpdateDefinition


if TYPE_CHECKING:
    from .._types import CurveUpdateDefinition, OptOverrides, Segments, OptTurns


class UpdateRequest(MixinRequest):
    """
    Update an existing cross currency curve request

    Parameters
    ----------
    curve_definition : CrossCurrencyCurveUpdateDefinition, optional

    overrides : list of OverrideBidAsk, optional

    segments : list of CrossCurrencyInstrumentsSegment

    turns : list of OverrideFxForwardTurn, optional

    """

    def __init__(
        self,
        curve_definition: "CurveUpdateDefinition" = None,
        overrides: "OptOverrides" = None,
        segments: "Segments" = None,
        turns: "OptTurns" = None,
    ) -> None:
        super().__init__()
        self.curve_definition = curve_definition
        self.overrides = try_copy_to_list(overrides)
        self.segments = try_copy_to_list(segments)
        self.turns = try_copy_to_list(turns)

    @property
    def curve_definition(self):
        """
        :return: object CrossCurrencyCurveUpdateDefinition
        """
        return self._get_object_parameter(
            CrossCurrencyCurveUpdateDefinition, "curveDefinition"
        )

    @curve_definition.setter
    def curve_definition(self, value):
        self._set_object_parameter(
            CrossCurrencyCurveUpdateDefinition, "curveDefinition", value
        )



import pandas as pd
from dataclasses import dataclass
from dateutil import parser

from pandas.tseries.holiday import AbstractHolidayCalendar, nearest_workday, Holiday
from collections import defaultdict
from typing import List, Optional, Dict

from ..._types import OptDateTime
from ...content.ipa.dates_and_calendars.holidays import Definition


class HolidayCalendar(AbstractHolidayCalendar):
    """
    Custom Holiday Calendar
    """

    def __init__(self, custom_holidays: list):
        super().__init__()
        self._custom_holidays = custom_holidays

    @property
    def rules(self) -> List[Holiday]:
        # rules are compatible with pandas
        return [
            self._create_rule(hol["date"], hol["name"]) for hol in self._custom_holidays
        ]

    @property
    def rules_list(self) -> List[dict]:
        # for easy access to all holidays the list of holidays is exposed
        return [
            self._create_holiday(hol["date"], hol["name"])
            for hol in self._custom_holidays
        ]

    @property
    def offset(self):
        return pd.offsets.CustomBusinessDay(calendar=self)

    @staticmethod
    def _create_holiday(date, name):
        date = parser.parse(date)
        return {"date": date, "name": name}

    @staticmethod
    def _create_rule(date, name, observance=None):
        # create a single rule from a holiday
        date = parser.parse(date)
        year, month, day = date.year, date.month, date.day
        observance = observance or nearest_workday
        a = Holiday(name=name, year=year, month=month, day=day, observance=observance)
        return a


def holidays_per_calendar(response) -> dict:
    def extract_holiday(holiday_info):
        return {
            "date": holiday_info.get("date"),
            "name": holiday_info.get("names")[0]["name"],
        }

    _holidays_per_calendar = defaultdict(list)
    for holiday in response:
        for cldr in holiday.get("calendars"):
            _holidays_per_calendar[cldr].append(extract_holiday(holiday))
    return _holidays_per_calendar


@dataclass
class HolidaysResponse:
    df: pd.DataFrame
    calendars: Dict[str, HolidayCalendar]
    offset: pd.offsets.CustomBusinessDay


def holidays(
    start_date: "OptDateTime" = None,
    end_date: "OptDateTime" = None,
    calendars: Optional[List[str]] = None,
    currencies: Optional[List[str]] = None,
) -> HolidaysResponse:
    """
    With this function you can get the holidays between
    a start date and an end date using one or many calendars.

    Parameters
    ----------
    start_date: str or datetime or timedelta, optional
        Start date of calculation.
    end_date: str or datetime or timedelta, optional
        End date of calculation.
    calendars: list of str, optional
        Calendars to use the date for working day or weekend.
        Optional if currencies is provided.
    currencies: list of str, optional
        Currencies to use the date for working day or weekend.
        Optional if calendars is provided.

    Returns
    -------
    HolidaysResponse
        HolidaysResponse object with dataframe, requested calendars and offset

    Examples
    --------
     >>> import datetime
     >>> import refinitiv.data as rd
     >>>
     >>> rd.open_session("platform.default")
     >>>
     >>> holidays_response = rd.dates_and_calendars.holidays(
     >>>    start_date=datetime.datetime(2019, 11, 20),
     >>>    end_date=datetime.datetime(2020, 2, 20),
     >>>    calendars=["EMU", "GER"],
     >>>)
    """

    holiday_outputs = ["Date", "Names", "Calendars", "Countries"]

    response = Definition(
        start_date=start_date,
        end_date=end_date,
        calendars=calendars,
        currencies=currencies,
        holiday_outputs=holiday_outputs,
    ).get_data()

    _holidays_per_calendar = holidays_per_calendar(
        response.data.raw[0].get("holidays", [])
    )

    holiday_calendars = {}
    for calendar, holidays_ in _holidays_per_calendar.items():
        holiday_calendars[calendar] = HolidayCalendar(holidays_)

    all_calendar_holidays = []
    for items in _holidays_per_calendar.values():
        for item in items:
            flag = True
            for holiday in all_calendar_holidays:
                if holiday["date"] == item["date"]:
                    flag = False
                    break

            if flag:
                all_calendar_holidays.append(item)

    global_holiday_calendar = HolidayCalendar(all_calendar_holidays)

    response = HolidaysResponse(
        response.data.df, holiday_calendars, global_holiday_calendar.offset
    )

    return response



import allure
import pytest

from refinitiv.data.content import filings
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.filings.conftest import (
    check_file_downloaded_response,
    check_file_is_downloaded,
    get_callback,
)
from tests.integration.helpers import (
    get_async_response_from_definition,
    check_response_status,
    check_non_empty_response_data,
    check_response_dataframe_contains_columns_names,
)


@allure.suite("Filings Retrieval")
@allure.feature("Filings Retrieval")
@allure.severity(allure.severity_level.CRITICAL)
class TestFilingsRetrieval:
    @allure.title("Create filings retrieval definition with valid params and get data")
    @pytest.mark.parametrize(
        "filename,dcn,doc_id,filing_id,expected_titles",
        [
            (
                "ecpfilings_34359955599_pdf",
                None,
                None,
                None,
                ["Filename", "SignedURL", "MimeType"],
            ),
            (None, "cr00329072", None, None, ["Filename", "SignedURL", "MimeType"]),
            (None, None, "49612437", None, ["Filename", "SignedURL", "MimeType"]),
            (None, None, None, "34359955599", ["Filename", "SignedURL", "MimeType"]),
        ],
    )
    @pytest.mark.caseid("C43473347")
    @pytest.mark.smoke
    def test_create_filings_retrieval_definition_with_valid_params_and_get_data(
        self, filename, dcn, doc_id, filing_id, expected_titles, open_platform_session
    ):
        response = filings.retrieval.Definition(
            filename=filename, dcn=dcn, doc_id=doc_id, filing_id=filing_id
        ).get_data()

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_response_dataframe_contains_columns_names(response, expected_titles)
        check_non_empty_response_data(response)

    @allure.title("Create filings retrieval definition and download file")
    @pytest.mark.parametrize(
        "filename",
        [
            "ecpfilings_34359955599_pdf",
        ],
    )
    @pytest.mark.caseid("C43473348")
    @pytest.mark.smoke
    def test_create_filings_retrieval_definition_and_download_file(
        self, filename, open_platform_session
    ):
        response = filings.retrieval.Definition(filename=filename).get_data()

        file = response.data.files
        file_download_response = file[0].download(path="./")
        expected_file_name = response.data.df["Filename"][0]

        check_file_downloaded_response(file_download_response, expected_file_name)
        check_file_is_downloaded(expected_file_name)

    @allure.title("Create filings retrieval definition and download all files")
    @pytest.mark.parametrize(
        "doc_id",
        [
            "49612437",
        ],
    )
    @pytest.mark.caseid("C43473350")
    @pytest.mark.smoke
    def test_create_filings_retrieval_definition_and_download_all_files(
        self, doc_id, open_platform_session
    ):
        response = filings.retrieval.Definition(doc_id=doc_id).get_data()

        files = response.data.files
        files.download()

        expected_file_name = response.data.df["Filename"].values
        for filename in expected_file_name:
            check_file_is_downloaded(filename)

    @allure.title("Create filings retrieval definition with more than one parameter")
    @pytest.mark.parametrize(
        "doc_id,filename",
        [("49612437", "file")],
    )
    @pytest.mark.caseid("C43473351")
    @pytest.mark.smoke
    def test_create_filings_retrieval_definition_with_more_than_one_parameter(
        self, doc_id, filename, open_platform_session
    ):
        with pytest.raises(ValueError) as error:
            filings.retrieval.Definition(doc_id=doc_id, filename=filename).get_data()

        assert (
            str(error.value)
            == "Only one of filename, dcn, doc_id or filing_id, can be used in a Definition"
        )

    @allure.title(
        "Create filings retrieval definition and download all files asynchronously"
    )
    @pytest.mark.parametrize(
        "doc_id,expected_titles",
        [("49612437", ["Filename", "SignedURL", "MimeType"])],
    )
    @pytest.mark.caseid("C43473352")
    @pytest.mark.smoke
    async def test_create_filings_retrieval_definition_and_download_all_files_asynchronously(
        self, doc_id, expected_titles, open_platform_session_async
    ):
        response = await get_async_response_from_definition(
            filings.retrieval.Definition(doc_id=doc_id)
        )

        files = response.data.files
        await files.download_async(callback=get_callback)

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_response_dataframe_contains_columns_names(response, expected_titles)
        expected_file_name = response.data.df["Filename"].values
        for filename in expected_file_name:
            check_file_is_downloaded(filename)



from datetime import datetime, timedelta
from typing import List, Optional, Union, TYPE_CHECKING

from numpy import iterable

from ._add_periods_data_provider import AddedPeriod, AddedPeriods
from .._base_request_items import StartDateBase
from ..._enums import DateMovingConvention, EndOfMonthConvention, HolidayOutputs
from ...._content_provider_layer import ContentUsageLoggerMixin
from ....._content_type import ContentType
from ....._tools import create_repr, make_enum_arg_parser, try_copy_to_list
from .....delivery._data._data_provider import DataProviderLayer, BaseResponse

if TYPE_CHECKING:
    from ....._types import ExtendedParams, OptDateTime, OptStrStrs
    from ..._enums._holiday_outupts import OptHolidayOutputs

holiday_outputs_arg_parser = make_enum_arg_parser(HolidayOutputs)


class DatePeriodsRequestItem(StartDateBase):
    def __init__(
        self,
        start_date: Union[str, datetime, timedelta] = None,
        period: str = None,
        calendars: Optional[List[str]] = None,
        currencies: Optional[List[str]] = None,
        tag: Optional[str] = None,
        date_moving_convention: Optional[DateMovingConvention] = None,
        end_of_month_convention: Optional[EndOfMonthConvention] = None,
        holiday_outputs: Optional[List[HolidayOutputs]] = None,
    ):
        super().__init__(start_date)
        self.tag = tag
        self.period = period
        self.calendars = calendars
        self.currencies = currencies
        self.date_moving_convention = date_moving_convention
        self.end_of_month_convention = end_of_month_convention
        self.holiday_outputs = holiday_outputs

    @property
    def tag(self):
        """
        :return: str
        """
        return self._get_parameter("tag")

    @tag.setter
    def tag(self, value):
        self._set_parameter("tag", value)

    @property
    def period(self):
        """
        :return: str
        """
        return self._get_parameter("period")

    @period.setter
    def period(self, value):
        self._set_parameter("period", value)

    @property
    def calendars(self):
        """
        :return: list
        """
        return self._get_parameter("calendars")

    @calendars.setter
    def calendars(self, value):
        self._set_parameter("calendars", value)

    @property
    def currencies(self):
        """
        :return: list
        """
        return self._get_parameter("currencies")

    @currencies.setter
    def currencies(self, value):
        self._set_parameter("currencies", value)

    @property
    def date_moving_convention(self):
        """
        :return: DateMovingConvention
        """
        return self._get_enum_parameter(DateMovingConvention, "dateMovingConvention")

    @date_moving_convention.setter
    def date_moving_convention(self, value):
        self._set_enum_parameter(DateMovingConvention, "dateMovingConvention", value)

    @property
    def end_of_month_convention(self):
        """
        :return: EndOfMonthConvention
        """
        return self._get_enum_parameter(EndOfMonthConvention, "endOfMonthConvention")

    @end_of_month_convention.setter
    def end_of_month_convention(self, value):
        self._set_enum_parameter(EndOfMonthConvention, "endOfMonthConvention", value)

    @property
    def holiday_outputs(self):
        """
        :return: list
        """
        return self._get_list_of_enums(HolidayOutputs, "holidayOutputs")

    @holiday_outputs.setter
    def holiday_outputs(self, value):
        self._set_list_of_enums(HolidayOutputs, "holidayOutputs", value)


class Definition(
    ContentUsageLoggerMixin[BaseResponse[AddedPeriod]],
    DataProviderLayer[BaseResponse[AddedPeriod]],
):
    """
    Add periods definition object

    Parameters
    ----------
        start_date: str or datetime or timedelta
            Start date of calculation.
        period: str
            String representing the tenor.
        calendars: list of str, optional
            Calendars to use the date for working day or weekend.
            Optional if currencies is provided.
        currencies: list of str, optional
            Currencies to use the date for working day or weekend.
            Optional if calendars is provided.
        tag: str, optional
            Reference tag to map particular response in payload.
        date_moving_convention : DateMovingConvention or str, optional
            The method to adjust dates.
        end_of_month_convention : EndOfMonthConvention or str, optional
            End of month convention.
        holiday_outputs : HolidayOutputs or list of str, optional
            In case if test date is holiday you may request additional information about the holiday.
            Possible options are: Date, Names, Calendars, Countries
        extended_params : dict, optional
            If necessary other parameters.

    Methods
    -------
    get_data(session=None, on_response=None, **kwargs)
        Returns a response to the data platform
    get_data_async(session=None, on_response=None, **kwargs)
        Returns a response asynchronously to the data platform

    Examples
    --------
     >>> from refinitiv.data.content.ipa.dates_and_calendars import add_periods
     >>> definition = add_periods.Definition(
     ...     start_date="2020-01-01",
     ...     period="4D",
     ...     calendars=["BAR", "KOR", "JAP"],
     ...     currencies=["USD"],
     ...     tag="my request",
     ...     date_moving_convention=add_periods.DateMovingConvention.NEXT_BUSINESS_DAY,
     ...     end_of_month_convention=add_periods.EndOfMonthConvention.LAST,
     ...     holiday_outputs=["Date", "Calendars", "Names"]
     ... )
     >>> response = definition.get_data()

     Using get_data_async
     >>> import asyncio
     >>> task = definition.get_data_async()
     >>> response = asyncio.run(task)
    """

    _USAGE_CLS_NAME = "IPA.DatesAndCalendars.AddPeriodsDefinition"

    def __init__(
        self,
        start_date: "OptDateTime" = None,
        period: str = None,
        calendars: "OptStrStrs" = None,
        currencies: "OptStrStrs" = None,
        tag: Optional[str] = None,
        date_moving_convention: Union[DateMovingConvention, str, None] = None,
        end_of_month_convention: Union[EndOfMonthConvention, str, None] = None,
        holiday_outputs: "OptHolidayOutputs" = None,
        extended_params: "ExtendedParams" = None,
    ):
        self.extended_params = extended_params
        calendars = try_copy_to_list(calendars)
        currencies = try_copy_to_list(currencies)
        if holiday_outputs:
            holiday_outputs = try_copy_to_list(holiday_outputs)
            holiday_outputs = holiday_outputs_arg_parser.get_list(holiday_outputs)

        self.request_item = DatePeriodsRequestItem(
            start_date=start_date,
            period=period,
            calendars=calendars,
            currencies=currencies,
            tag=tag,
            date_moving_convention=date_moving_convention,
            end_of_month_convention=end_of_month_convention,
            holiday_outputs=holiday_outputs,
        )

        super().__init__(
            data_type=ContentType.DATES_AND_CALENDARS_ADD_PERIODS,
            universe=[self.request_item],
            extended_params=extended_params,
        )

    def __repr__(self):
        return create_repr(self)


DefnDefns = Union[List[Definition], Definition]


class Definitions(
    ContentUsageLoggerMixin[BaseResponse[AddedPeriods]],
    DataProviderLayer[BaseResponse[AddedPeriods]],
):
    """
    Add periods definitions object

    Parameters
    ----------
    universe: Definition or list of Definition objects
        List of initialized Definition objects to retrieve data.

    Methods
    -------
    get_data(session=None, on_response=None, **kwargs)
        Returns a response to the data platform
    get_data_async(session=None, on_response=None, **kwargs)
        Returns a response asynchronously to the data platform

    Examples
    --------
     >>> import datetime
     >>> from refinitiv.data.content.ipa.dates_and_calendars import add_periods
     >>>
     >>> first_definition = add_periods.Definition(
     ...     tag="first",
     ...     start_date="2020-01-01",
     ...     period="4D",
     ...     calendars=["BAR", "KOR", "JAP"],
     ...     currencies=["USD"],
     ...     date_moving_convention=add_periods.DateMovingConvention.NEXT_BUSINESS_DAY,
     ...     end_of_month_convention=add_periods.EndOfMonthConvention.LAST,
     ...     holiday_outputs=["Date", "Calendars", "Names"]
     ... )
     >>> second_definition = add_periods.Definition(
     ...     tag="second",
     ...     start_date="2018-01-01",
     ...     period="4D",
     ...     calendars=["BAR", "JAP"],
     ...     currencies=["USD"],
     ...     date_moving_convention=add_periods.DateMovingConvention.NEXT_BUSINESS_DAY,
     ...     end_of_month_convention=add_periods.EndOfMonthConvention.LAST28,
     ...     holiday_outputs=[add_periods.HolidayOutputs.DATE, add_periods.HolidayOutputs.NAMES]
     ... )
     >>> response = add_periods.Definitions([first_definition, second_definition]).get_data()

    Using get_data_async
     >>> import asyncio
     >>> task = add_periods.Definitions([first_definition, second_definition]).get_data_async()
     >>> response = asyncio.run(task)
    """

    _USAGE_CLS_NAME = "IPA.DatesAndCalendars.AddPeriodsDefinitions"

    def __init__(self, universe: "DefnDefns"):
        universe = try_copy_to_list(universe)
        if not iterable(universe):
            universe = [universe]

        request_items = []
        extended_params = []
        for item in universe:
            request_items.append(item.request_item)
            extended_params.append(item.extended_params)

        super().__init__(
            data_type=ContentType.DATES_AND_CALENDARS_ADD_PERIODS,
            universe=request_items,
            extended_params=extended_params,
        )

    def __repr__(self):
        return create_repr(self)



import time
from queue import Empty, Queue
from threading import Thread, Lock
from typing import List, Set, Optional, Type, Tuple

from ._abstract_logger import AbstractUsageLogger
from ._utils import LogRecord, RecordData, LoggerConfig
from ._filter_types import FilterType
from .._configure import get_config


class UsageLoggerProxy(Thread):
    """
    This class is a wrapper around a threading.Thread that runs a
    logging thread. It is used to log usage data from the RD Library.

    Parameters
    ----------
    batch_size : int
        The maximum number of records to send to the loggers at once.
    flush_timeout : int
        The maximum time to wait before flushing the batch buffer.
    logging_enabled : bool
        Whether to enable logging.

    Examples
    --------
    >>> from refinitiv.data.usage_collection import get_usage_logger
    >>> class UserLogger(AbstractUsageLogger):
    >>>     def __init__(self, *args, **kwargs):
    >>>         ...
    >>>     def log(self, records: List[LogRecord]) -> None:
    >>>         ...
    >>>     def close(self) -> None:
    >>>         ...
    >>>
    >>> logger = get_usage_logger()
    >>> logger.add_logger(UserLogger)
    >>> logger.start()
    >>> ...
    >>> logger.join()
    """

    def __init__(
        self,
        batch_size: int = 100,
        flush_timeout: int = 10,
        logging_enabled: bool = True,
    ) -> None:
        super().__init__(name="UsageLoggerThread", daemon=True)
        self._queue: Queue = Queue()
        self._loggers: List[LoggerConfig] = []
        self._lock = Lock()
        self._logger_instances: List[Tuple[AbstractUsageLogger, Set[FilterType]]] = []
        self._batch_buffer: List[LogRecord] = []
        self._batch_size: int = batch_size
        self._flush_timeout: int = flush_timeout
        self._logging_enabled = logging_enabled
        self._last_flush: int = 0

    @property
    def logging_enabled(self):
        return self._logging_enabled

    @property
    def queue(self):
        return self._queue

    @staticmethod
    def _filter_batch(
        batch: List[LogRecord], logger_filter: Optional[Set[FilterType]]
    ) -> List[LogRecord]:
        """
        Filter the batch by the given filter.
        Parameters
        ----------
        batch : List[LogRecord]
        logger_filter : Optional[Set[FilterType]]

        Returns
        -------

        """
        return [
            record
            for record in batch
            if not logger_filter or (logger_filter & record.filter)
        ]

    def flush(self) -> None:
        """
        Flush the batch buffer to the loggers.
        Returns
        -------

        """
        if self._batch_buffer:
            # For each registered logger send the whole buffer
            for _logger, _filter in self._logger_instances:
                _logger.log(self._filter_batch(self._batch_buffer, _filter))
            self._batch_buffer.clear()
            self._last_flush = time.monotonic()

    def _update_loggers(self) -> None:
        """
        Update the list of logger instances.
        Returns
        -------

        """
        with self._lock:
            for logger_config in self._loggers:
                self._logger_instances.append(
                    (
                        logger_config.logger_type(
                            *logger_config.args, **logger_config.kwargs
                        ),
                        logger_config.filters,
                    )
                )
            self._loggers = []

    def run(self) -> None:
        self._last_flush = time.monotonic()

        while True:
            if len(self._loggers) > 0:
                self._update_loggers()
            try:
                record = self._queue.get(
                    timeout=max(
                        0.1, self._flush_timeout - (time.monotonic() - self._last_flush)
                    )
                )
            except Empty:
                self.flush()
                continue
            if record is None:
                break
            self._batch_buffer.append(record)
            if len(self._batch_buffer) >= self._batch_size:
                self.flush()
        self.flush()
        for logger, _ in self._logger_instances:
            logger.close()

    def start(self) -> None:
        if self._logging_enabled:
            super().start()

    def add_logger(
        self,
        logger: Type[AbstractUsageLogger],
        _filter: Set[FilterType],
        *args,
        **kwargs,
    ) -> None:
        """
        Add a logger to the logging thread.
        """
        if self._logging_enabled:
            if not issubclass(logger, AbstractUsageLogger):
                raise ValueError("Logger must be a subclass of UsageLogger")
            with self._lock:
                self._loggers.append(LoggerConfig(logger, args, kwargs, _filter))
        else:
            raise RuntimeError(
                "Tried to add a logger to a disabled logger thread. "
                "Check session config"
            )

    def log(self, record: LogRecord) -> None:
        if self._logging_enabled:
            self.queue.put(record)

    def log_func(
        self,
        name: str,
        func_path: str,
        args: tuple = None,
        kwargs: dict = None,
        result: object = None,
        desc: Set[FilterType] = None,
    ) -> None:
        if self._logging_enabled:
            if args is None:
                args = ()
            if kwargs is None:
                kwargs = {}
            if desc is None:
                desc = set()
            self.log(
                LogRecord(name, func_path, RecordData(args, kwargs, result), desc),
            )

    def join(self, timeout: Optional[float] = None) -> None:
        """
        Stop the logger thread.
        Parameters
        ----------
        timeout : Optional[float]

        Returns
        -------

        """
        if self._logging_enabled:
            self.queue.put(None)
            try:
                super().join(timeout)
            except RuntimeError:
                pass
            self._logging_enabled = False


usage_logger: Optional[UsageLoggerProxy] = None


def get_usage_logger() -> UsageLoggerProxy:
    global usage_logger
    if usage_logger is None:
        logging_enabled = get_config().get_bool("usage_logger.enabled")
        usage_logger = UsageLoggerProxy(logging_enabled=logging_enabled)
        usage_logger.start()
    return usage_logger



from types import SimpleNamespace

import pytest

from refinitiv.data._tools import (
    create_repr,
    get_new_path,
)
from tests.unit.conftest import kwargs

PREFIX_PATH = "test"


def mock_object(path_module=f"{PREFIX_PATH}._data.public_path"):
    obj = SimpleNamespace()
    obj.__module__ = path_module
    return obj


@pytest.mark.parametrize(
    "input_value, expected",
    [
        ("", "<.Definition object at"),
        ("._data.", "<.Definition object at"),
        ("test", "<test.Definition object at"),
        ("test.public_path", "<test.public_path.Definition object at"),
    ],
)
def test_create_repr_without_correct_module_path(input_value, expected):
    # given
    mock_obj = mock_object(input_value)

    # when
    repr_str = create_repr(mock_obj)

    # then
    assert repr_str.startswith(expected), f"mismatch of initial strings {repr_str}"


def test_create_repr_not_validate_first_arg():
    # when
    with pytest.raises(AttributeError):
        create_repr({})


@pytest.mark.parametrize(
    "input_args, expected",
    [
        ({}, f"<{PREFIX_PATH}.Definition object at"),
        (
            kwargs(middle_path="context"),
            f"<{PREFIX_PATH}.context.Definition object at",
        ),
        (
            kwargs(middle_path="context", class_name="Stream"),
            f"<{PREFIX_PATH}.context.Stream object at",
        ),
    ],
)
def test_create_repr_with_validate_other_args(input_args, expected):
    # given
    mock_obj = mock_object()

    # when
    repr_str = create_repr(mock_obj, **input_args)

    # then
    assert repr_str.startswith(expected), f"mismatch of initial strings {repr_str}"


def test_create_repr_with_validate_content():
    # given
    mock_obj = mock_object()

    # when
    repr_str = create_repr(mock_obj, content="content_test")

    # then
    assert repr_str.startswith(f"<{PREFIX_PATH}.Definition object at")
    assert repr_str.endswith(" content_test>")


def test_create_repr_not_validate_content():
    # given
    mock_obj = mock_object()

    # when
    repr_str = create_repr(mock_obj, content="")

    # then
    assert repr_str.startswith(f"<{PREFIX_PATH}.Definition object at")
    assert repr_str.endswith(" >") is False


@pytest.mark.parametrize(
    "input_value, expected_value",
    [
        (
            kwargs(
                module_path="refinitiv.data._definition",
                class_name="Definition",
                middle="content",
            ),
            "refinitiv.data.content.Definition",
        ),
        (
            kwargs(
                module_path="refinitiv.data._definition",
                class_name="Definition",
            ),
            "refinitiv.data.Definition",
        ),
        (
            kwargs(
                module_path="refinitiv.data._definition",
                class_name="Definition",
                middle="content.pricing",
            ),
            "refinitiv.data.content.pricing.Definition",
        ),
    ],
)
def test_get_new_path(input_value, expected_value):
    # when
    testing_value = get_new_path(**input_value)

    # then
    assert testing_value == expected_value



from concurrent.futures import Future, wait
from typing import Any, Callable, Dict, List, TYPE_CHECKING, Tuple

from ._chain_records import ChainRecords
from ._display_template import disp_tmpl_to_num_summary_links
from ....delivery._stream import (
    OMMStreamListener,
    StreamEvent,
    StreamStateManager,
)

if TYPE_CHECKING:
    from ...._types import ExtendedParams, OptBool, OptInt, OptStr
    from ....delivery._stream import _OMMStream
    from ...._core.session import Session


class StreamingChainEvent:
    ADD = "streaming_chain_add_event"
    REMOVE = "streaming_chain_remove_event"


class StreamingChain(StreamStateManager, OMMStreamListener["StreamingChain"]):
    def __init__(
        self,
        name: str,
        session: "Session",
        service: "OptStr" = None,
        skip_summary_links: "OptBool" = True,
        skip_empty: "OptBool" = True,
        override_summary_links: "OptInt" = None,
        extended_params: "ExtendedParams" = None,
    ):
        self._session = session

        StreamStateManager.__init__(self, logger=self._session.logger())
        OMMStreamListener.__init__(
            self, logger=self._session.logger(), add_originator=False
        )

        self.name: str = name
        self._service = service
        self._skip_summary_links = skip_summary_links
        self._skip_empty = skip_empty
        self._override_summary_links = override_summary_links
        self._chain_records: ChainRecords = ChainRecords(
            self._session, service=service, extended_params=extended_params
        )
        self._chain_records.on_refresh(self._on_stream_refresh)
        self._chain_records.on_status(self._on_stream_status)
        self._chain_records.on_update(self._on_stream_update)
        self._chain_records.on_error(self._on_stream_error)

        self._constituents = None
        self._complete_future: Future = Future()
        self._chain_record_name_to_offset: Dict[str, int] = {self.name: 0}
        self._update_messages: List[Tuple[str, dict]] = []

    @property
    def is_chain(self) -> bool:
        wait([self._complete_future])
        is_chain = self._chain_records.has(self.name)
        return is_chain

    @property
    def num_summary_links(self) -> int:
        if self._override_summary_links:
            return self._override_summary_links
        else:
            disp_tmpl = self._chain_records.display_template
            return disp_tmpl_to_num_summary_links.get(disp_tmpl, None)

    @property
    def summary_links(self) -> List[str]:
        if not self.is_chain:
            return []

        summary_links = self._constituents[: self.num_summary_links]
        return [
            summary_link
            for summary_link in summary_links
            if not self._skip_empty or summary_link is not None
        ]

    @property
    def display_name(self) -> str:
        if not self.is_chain:
            return ""

        return self._chain_records.get_display_name(self.name)

    @property
    def session(self) -> "Session":
        return self._session

    @session.setter
    def session(self, session: "Session"):
        if self._session != session and not self.is_open:
            self._session = session
            OMMStreamListener.init_logger(self, self._session.logger())
            self._chain_records.session = self._session

    def get_constituents(self) -> List[str]:
        if not self.is_chain:
            return []

        num_summary_links = self.num_summary_links
        if self._skip_summary_links:
            constituents = self._constituents[num_summary_links:]
        else:
            constituents = self._constituents[:]

        return [
            constituent
            for constituent in constituents
            if not self._skip_empty or constituent is not None
        ]

    def _do_open(self, *args, with_updates=True):
        self._chain_records.open(with_updates=with_updates)
        self._constituents = []
        offset = 0
        name = self.name
        while name:
            if not self._chain_records.has_stream(name):
                self._chain_records.add(name)
                self._chain_records.open_stream(name, with_updates)

            self._chain_records.wait_refresh(name)

            if self._chain_records.is_stream_close(name):
                self._complete_future.set_result(True)
                break

            chain_record = self._chain_records.get_record(name)
            name = chain_record
            if chain_record:
                name = chain_record.next_chain_record_name

                for i, constituent in enumerate(chain_record.constituents):
                    self._append_constituent(offset + i, constituent)

                offset += chain_record.num_constituents or 0

                if name:
                    self._chain_record_name_to_offset[name] = offset

        if not self._complete_future.done():
            self._complete_future.set_result(True)

        self.dispatch_complete(self, self.get_constituents())
        self._process_remaining_update_messages()

    def _do_close(self, *args, **kwargs):
        self._chain_records.close()

    def _append_constituent(self, index: int, constituent: str):
        self._constituents.append(constituent)
        self.dispatch_add(constituent, index)

    def _remove_constituent(self, index: int, constituent: str):
        self._constituents.pop(index)
        self.dispatch_remove(constituent, index)

    def _update_constituent(
        self, index: int, old_constituent: str, new_constituent: str
    ):
        self._constituents[index] = new_constituent
        self.dispatch_update(self, new_constituent, old_constituent, index)

    def _do_on_stream_status(self, originator: ChainRecords, message: dict, *_) -> Any:
        state = message.get("State", {})
        stream_state = state.get("Stream")
        if stream_state == "Closed":
            self._chain_records.close()

        return message

    def _on_stream_update(
        self, originator: ChainRecords, stream: "_OMMStream", message: dict, *args
    ) -> Any:
        if not self._complete_future.done():
            self._update_messages.append((stream.name, message))
            self._warning(
                "StreamingChain :: "
                "waiting to update because chain decode does not completed."
            )
            return

        self._process_remaining_update_messages()
        self._update_chain_record(stream.name, message)

    def _process_remaining_update_messages(self):
        self._debug(f"{self._classname} starts process remaining update messages")
        while True:
            try:
                (name, message) = self._update_messages.pop(0)
            except IndexError:
                break

            self._update_chain_record(name, message)

    def _update_chain_record(self, name: str, message: dict):
        self._debug(f"{self._classname} updates chain record, name={name}")
        fields = message.get("Fields", [])
        chain_record = self._chain_records.get_record(name)

        if not chain_record:
            self._warning(
                f"StreamingChain :: "
                f"Skipping to update an invalid chain record = {name}."
            )
            return

        index_to_old_and_new_constituent = chain_record.update(fields)

        offset = self._chain_record_name_to_offset[name]
        for i, (old_c, new_c) in index_to_old_and_new_constituent.items():
            i = offset + i

            if old_c and new_c:
                self._update_constituent(i, old_c, new_c)

            elif not old_c and new_c:
                self._append_constituent(i, new_c)

            elif old_c and not new_c:
                self._remove_constituent(i, old_c)

    def dispatch_update(self, originator: "StreamingChain", *args):
        self._debug(f"{self._classname} dispatch_update {args}")
        self._emitter.emit(StreamEvent.UPDATE, *args)

    def dispatch_add(self, constituent, index):
        self._debug(f"{self._classname} dispatch_add {index, constituent}")
        self._emitter.emit(StreamingChainEvent.ADD, index, constituent)

    def dispatch_remove(self, constituent, index):
        self._debug(f"{self._classname} dispatch_remove {index, constituent}")
        self._emitter.emit(StreamingChainEvent.REMOVE, index, constituent)

    def on_add(self, func: Callable):
        self._on_event(StreamingChainEvent.ADD, func, halt_if_none=True)

    def on_remove(self, func):
        self._on_event(StreamingChainEvent.REMOVE, func, halt_if_none=True)



from refinitiv.data.content.ipa._curves._bond_curves._curves import (
    CreditCurveDefinition,
)
from refinitiv.data.content.ipa._curves._bond_curves._enums import (
    BusinessSector,
    CurveSubType,
    EconomicSector,
    Industry,
    IndustryGroup,
    IssuerType,
    MainConstituentAssetClass,
    Rating,
    RatingScaleSource,
    ReferenceEntityType,
    Seniority,
)


def test_credit_curve_definition():
    # given
    expected_dict = {
        "businessSector": "AcademicAndEducationalServices",
        "creditCurveTypeFallbackLogic": ["string1", "string2"],
        "curveSubType": "BondCarry",
        "curveTenors": ["string1", "string2"],
        "economicSector": "AcademicAndEducationalServices",
        "industry": "AdvancedMedicalEquipmentAndTechnology",
        "industryGroup": "AerospaceAndDefense",
        "issuerType": "Agency",
        "mainConstituentAssetClass": "Bond",
        "rating": "A",
        "ratingScaleSource": "DBRS",
        "referenceEntityType": "BondIsin",
        "seniority": "SeniorNonPreferred",
        "country": "string",
        "currency": "string",
        "id": "string",
        "name": "string",
        "referenceEntity": "string",
        "source": "string",
    }

    # when
    testing_obj = CreditCurveDefinition(
        business_sector=BusinessSector.ACADEMIC_AND_EDUCATIONAL_SERVICES,
        credit_curve_type_fallback_logic=["string1", "string2"],
        curve_sub_type=CurveSubType.BOND_CARRY,
        curve_tenors=["string1", "string2"],
        economic_sector=EconomicSector.ACADEMIC_AND_EDUCATIONAL_SERVICES,
        industry=Industry.ADVANCED_MEDICAL_EQUIPMENT_AND_TECHNOLOGY,
        industry_group=IndustryGroup.AEROSPACE_AND_DEFENSE,
        issuer_type=IssuerType.AGENCY,
        main_constituent_asset_class=MainConstituentAssetClass.BOND,
        rating=Rating.A,
        rating_scale_source=RatingScaleSource.DBRS,
        reference_entity_type=ReferenceEntityType.BOND_ISIN,
        seniority=Seniority.SENIOR_NON_PREFERRED,
        country="string",
        currency="string",
        id="string",
        name="string",
        reference_entity="string",
        source="string",
    )

    # then
    assert testing_obj.get_dict() == expected_dict

    assert (
        testing_obj.business_sector == BusinessSector.ACADEMIC_AND_EDUCATIONAL_SERVICES
    )
    assert testing_obj.credit_curve_type_fallback_logic == ["string1", "string2"]
    assert testing_obj.curve_sub_type == CurveSubType.BOND_CARRY
    assert testing_obj.curve_tenors == ["string1", "string2"]
    assert (
        testing_obj.economic_sector == EconomicSector.ACADEMIC_AND_EDUCATIONAL_SERVICES
    )
    assert testing_obj.industry == Industry.ADVANCED_MEDICAL_EQUIPMENT_AND_TECHNOLOGY
    assert testing_obj.industry_group == IndustryGroup.AEROSPACE_AND_DEFENSE
    assert testing_obj.issuer_type == IssuerType.AGENCY
    assert testing_obj.main_constituent_asset_class == MainConstituentAssetClass.BOND
    assert testing_obj.rating == Rating.A
    assert testing_obj.rating_scale_source == RatingScaleSource.DBRS
    assert testing_obj.reference_entity_type == ReferenceEntityType.BOND_ISIN
    assert testing_obj.seniority == Seniority.SENIOR_NON_PREFERRED
    assert testing_obj.country == "string"
    assert testing_obj.currency == "string"
    assert testing_obj.id == "string"
    assert testing_obj.name == "string"
    assert testing_obj.reference_entity == "string"
    assert testing_obj.source == "string"



zc_curves_universe = {
    "curveDefinition": {
        "currency": "CHF",
        "discountingTenor": "OIS",
        "name": "CHF LIBOR Swap ZC Curve",
    },
    "curveParameters": {"useSteps": True, "valuationDate": "2019-08-21"},
}



import warnings
from typing import Optional, List, Union, Callable, TYPE_CHECKING

from ._enums import CustomInstrumentTypes
from ._instrument_class import (
    create_instr_factory,
    CustomInstrumentFormula,
    CustomInstrumentBasket,
    CustomInstrumentUDC,
)
from ..._core.session import Session
from ._instrument_prop_classes import Basket, UDC
from ..._content_type import ContentType
from ..ipa.dates_and_calendars.holidays._holidays_data_provider import Holiday
from ...delivery._data._data_provider import DataProviderLayer, Response
from ...delivery._data._endpoint_data import RequestMethod

if TYPE_CHECKING:
    from ..._types import ExtendedParams


def delete(
    universe: str,
    extended_params: "ExtendedParams" = None,
    session: "Session" = None,
) -> Response:
    """
    universe : str
        Instrument symbol in the format "S)someSymbol.YOURUUID".
    extended_params : ExtendedParams, optional
        If necessary other parameters.
    session : Session, optional
        session=None. Means default session would be used

    Examples
    --------
    >>> from refinitiv.data.content.custom_instruments.manage import delete
    >>> response = delete("MyInstrument")
    """
    data_provider_layer = DataProviderLayer(
        data_type=ContentType.CUSTOM_INSTRUMENTS_INSTRUMENTS,
        universe=universe,
        extended_params=extended_params,
        method=RequestMethod.DELETE,
    )
    return data_provider_layer.get_data(session)


def get(
    universe: str, extended_params: "ExtendedParams" = None, session: "Session" = None
) -> Union[CustomInstrumentFormula, CustomInstrumentBasket, CustomInstrumentUDC]:
    """
    universe : str
        Instrument symbol in the format "S)someSymbol.YOURUUID".
    extended_params : ExtendedParams, optional
        If necessary other parameters.
    session : Session, optional
        session=None - means default session would be used

    Examples
    --------
    >>> from refinitiv.data.content.custom_instruments.manage import get
    >>> response = get("MyInstrument")
    """
    data_provider_layer = DataProviderLayer(
        data_type=ContentType.CUSTOM_INSTRUMENTS_INSTRUMENTS,
        universe=universe,
        method=RequestMethod.GET,
        extended_params=extended_params,
    )
    response = data_provider_layer.get_data(session=session)
    return create_instr_factory(response.data.raw, session=session)


# deprecated method
def create(
    symbol: str,
    formula: Optional[str] = None,
    basket: Optional[dict] = None,
    udc: Optional[dict] = None,
    instrument_name: Optional[str] = None,
    exchange_name: Optional[str] = None,
    currency: Optional[str] = None,
    time_zone: Optional[str] = None,
    holidays: Optional[List[Union[dict, Holiday]]] = None,
    description: Optional[str] = None,
    type_: Union[str, CustomInstrumentTypes] = CustomInstrumentTypes.Formula,
    extended_params: "ExtendedParams" = None,
    session: "Session" = None,
    on_response: Callable = None,
) -> Union[CustomInstrumentFormula, CustomInstrumentBasket, CustomInstrumentUDC]:
    """
    With this method you can create a CustomInstrumentFormula, CustomInstrumentBasket, CustomInstrumentUDC objects.

    Parameters
    ----------
    symbol: str
        Instrument symbol in the format "S)someSymbol.YOURUUID".
    formula : str
        Formula consisting of rics (fields can be specified by comma).
    basket : dict, Basket, optional
        Method of defining custom instruments, relying on a list of instruments and weights in order to build
        custom indices or other simple synthetic instruments
    udc : dict, UDC, optional
        User-Defined Continuations. Custom trading sessions, see sample format below.
    currency : str, optional
        3-letter code of the currency of the instrument, e.g. GBP.
    instrument_name : str, optional
        Human-readable name of the instrument. Maximum of 16 characters.
    exchange_name : str, optional
        4-letter code of the listing exchange.
    holidays : list[dict, Holiday], optional
        List of custom calendar definitions.
    time_zone : str, optional
        Time Series uses an odd custom 3-letter value for time zone IDs, e.g. "LON" for London.
    description : str, optional
        Free text field from the user to put any notes or text. Up to 1000 characters.
    type_ : str, CustomInstrumentTypes, optional
        Type of Synthetic Instrument - "formula", "basket","udc". Default value is "formula".
    extended_params : ExtendedParams, optional
        If necessary other parameters.
    session : Session, optional
        session=None - means default session would be used
    on_response : Callable, optional
        Callable object to process retrieved data

    Returns
    -------
        CustomInstrumentFormula

    Examples
    --------
    >>> from refinitiv.data.content.custom_instruments.manage import create_formula
    >>> import refinitiv.data.content.custom_instruments as ci
    >>> from refinitiv.data.content.ipa import dates_and_calendars
    >>> calendar_holiday = dates_and_calendars.holidays.Definition(
    ...     start_date="2015-08-24",
    ...     end_date="2018-09-24",
    ...     calendars=["UKR"],
    ...     holiday_outputs=["Date", "Names"],
    >>> ).get_data()
    ...
    >>> response = create_formula(
    ...     symbol="MyNewInstrument",
    ...     formula="EUR=*3",
    ...     holidays=[
    ...         *calendar_holiday.data.holidays,
    ...         ci.manage.Holiday(date="1991-08-24", name="Independence Day of Ukraine"),
    ...         {"date": "2022-12-18", "reason": "Hanukkah"},
    ...     ],
    >>> )
    """
    warnings.warn(
        "'create()' is legacy interface. "
        "Will be changed to 'create_formula()', 'create_basket()', 'create_udc()'",
    )
    data = _create(
        symbol,
        formula,
        basket,
        udc,
        instrument_name,
        exchange_name,
        currency,
        time_zone,
        holidays,
        description,
        type_,
        extended_params,
        session,
        on_response,
    )
    return create_instr_factory(data, session=session)


def _create(
    symbol: str,
    formula: Optional[str] = None,
    basket: Union[dict, Basket] = None,
    udc: Union[dict, UDC] = None,
    instrument_name: Optional[str] = None,
    exchange_name: Optional[str] = None,
    currency: Optional[str] = None,
    time_zone: Optional[str] = None,
    holidays: Optional[List[Union[dict, Holiday]]] = None,
    description: Optional[str] = None,
    type_: Union[str, CustomInstrumentTypes] = None,
    extended_params: "ExtendedParams" = None,
    session: "Session" = None,
    on_response: Callable = None,
) -> dict:
    data_provider_layer = DataProviderLayer(
        data_type=ContentType.CUSTOM_INSTRUMENTS_INSTRUMENTS,
        symbol=symbol,
        formula=formula,
        instrument_name=instrument_name,
        exchange_name=exchange_name,
        currency=currency,
        time_zone=time_zone,
        holidays=holidays,
        description=description,
        type_=type_,
        basket=basket,
        udc=udc,
        extended_params=extended_params,
        method=RequestMethod.POST,
    )
    response = data_provider_layer.get_data(session, on_response)
    return response.data.raw


def create_formula(
    symbol: str,
    formula: Optional[str] = None,
    currency: Optional[str] = None,
    instrument_name: Optional[str] = None,
    exchange_name: Optional[str] = None,
    holidays: Optional[List[Union[dict, Holiday]]] = None,
    time_zone: Optional[str] = None,
    description: Optional[str] = None,
    extended_params: "ExtendedParams" = None,
    session: "Session" = None,
    on_response: Callable = None,
) -> CustomInstrumentFormula:
    """
    With this method you can create a CustomInstrumentFormula object.

    Parameters
    ----------
    symbol: str
        Instrument symbol in the format "S)someSymbol.YOURUUID".
    formula : str
        Formula consisting of rics (fields can be specified by comma).
    currency : str, optional
        3-letter code of the currency of the instrument, e.g. GBP.
    instrument_name : str, optional
        Human-readable name of the instrument. Maximum of 16 characters.
    exchange_name : str, optional
        4-letter code of the listing exchange.
    holidays : list[dict, Holiday], optional
        List of custom calendar definitions.
    time_zone : str, optional
        Time Series uses an odd custom 3-letter value for time zone IDs, e.g. "LON" for London.
    description : str, optional
        Free text field from the user to put any notes or text. Up to 1000 characters.
    extended_params : ExtendedParams, optional
        If necessary other parameters.
    session : Session, optional
        session=None - means default session would be used
    on_response : Callable, optional
        Callable object to process retrieved data

    Returns
    -------
        CustomInstrumentFormula

    Examples
    --------
    >>> from refinitiv.data.content.custom_instruments.manage import create_formula
    >>> import refinitiv.data.content.custom_instruments as ci
    >>> response = create_formula(
    ...     symbol="MyNewInstrument",
    ...     formula="EUR=*3",
    ...     holidays=[
    ...         ci.manage.Holiday(date="1991-08-23", name="Independence Day of Ukraine"),
    ...         {"date": "2022-12-18", "reason": "Hanukkah"},
    ...     ],
    >>> )
    """
    data = _create(
        symbol=symbol,
        type_=CustomInstrumentTypes.Formula,
        formula=formula,
        currency=currency,
        instrument_name=instrument_name,
        exchange_name=exchange_name,
        holidays=holidays,
        time_zone=time_zone,
        description=description,
        extended_params=extended_params,
        session=session,
        on_response=on_response,
    )
    return CustomInstrumentFormula(data, session=session)


def create_basket(
    symbol: str,
    basket: Union[dict, Basket],
    currency: str,
    instrument_name: Optional[str] = None,
    exchange_name: Optional[str] = None,
    holidays: Optional[List[Union[dict, Holiday]]] = None,
    time_zone: Optional[str] = None,
    description: Optional[str] = None,
    extended_params: "ExtendedParams" = None,
    session: "Session" = None,
    on_response: Callable = None,
) -> CustomInstrumentBasket:
    """
    With this method you can create a CustomInstrumentBasket object.

    Parameters
    ----------
    symbol: str
        Instrument symbol in the format "S)someSymbol.YOURUUID".
    basket : dict, Basket
        Method of defining custom instruments, relying on a list of instruments and weights in order to build
        custom indices or other simple synthetic instruments
    currency : str
        3-letter code of the currency of the instrument, e.g. GBP.
    instrument_name : str, optional
        Human-readable name of the instrument. Maximum of 16 characters.
    exchange_name : str, optional
        4-letter code of the listing exchange.
    holidays : list[dict, Holiday], optional
        List of custom calendar definitions.
    time_zone : str, optional
        Time Series uses an odd custom 3-letter value for time zone IDs, e.g. "LON" for London.
    description : str, optional
        Free text field from the user to put any notes or text. Up to 1000 characters.
    extended_params : ExtendedParams, optional
        If necessary other parameters.
    session : Session, optional
        session=None - means default session would be used
    on_response : Callable, optional
        Callable object to process retrieved data

    Returns
    -------
        CustomInstrumentBasket

    Examples
    --------
    >>> from refinitiv.data.content.custom_instruments.manage import create_basket
    >>> import refinitiv.data.content.custom_instruments as ci
    >>> import datetime
    ... response = create_basket(
    ...        symbol="MyBasketInstrument",
    ...        holidays=[
    ...            ci.manage.Holiday(date="1991-10-24", name="Labour Day"),
    ...            ci.manage.Holiday(date=datetime.date(2021, 8, 24), name="Independence Day of Ukraine"),
    ...            ci.manage.Holiday(date=datetime.timedelta(days=-30), name="Alaska Day"),
    ...            {"date": "2022-04-23", "reason": "Shakespeare Day"},
    ...        ],
    ...        basket=Basket(
    ...            constituents=[
    ...                ci.manage.Constituent(ric="LSEG.L", weight=50),
    ...                ci.manage.Constituent(ric="EPAM.N", weight=50),
    ...            ],
    ...            normalize_by_weight=True,
    ...        ),
    ...        currency="USD",
    ...    )
    """

    data = _create(
        symbol=symbol,
        type_=CustomInstrumentTypes.Basket,
        basket=basket,
        currency=currency,
        instrument_name=instrument_name,
        exchange_name=exchange_name,
        holidays=holidays,
        time_zone=time_zone,
        description=description,
        extended_params=extended_params,
        session=session,
        on_response=on_response,
    )
    return CustomInstrumentBasket(data, session=session)


def create_udc(
    symbol: str,
    udc: Union[dict, UDC],
    currency: Optional[str] = None,
    instrument_name: Optional[str] = None,
    exchange_name: Optional[str] = None,
    holidays: Optional[List[Union[dict, Holiday]]] = None,
    time_zone: Optional[str] = None,
    description: Optional[str] = None,
    extended_params: "ExtendedParams" = None,
    session: "Session" = None,
    on_response: Callable = None,
) -> CustomInstrumentUDC:
    """
    With this method you can create a CustomInstrumentUDC object.

    Parameters
    ----------
    symbol: str
        Instrument symbol in the format "S)someSymbol.YOURUUID".
    udc : dict, UDC
        User-Defined Continuations. Custom trading sessions, see sample format below.
    currency : str, optional
        3-letter code of the currency of the instrument, e.g. GBP.
    instrument_name : str, optional
        Human-readable name of the instrument. Maximum of 16 characters.
    exchange_name : str, optional
        4-letter code of the listing exchange.
    holidays : list[dict, Holiday], optional
        List of custom calendar definitions.
    time_zone : str, optional
        Time Series uses an odd custom 3-letter value for time zone IDs, e.g. "LON" for London.
    description : str, optional
        Free text field from the user to put any notes or text. Up to 1000 characters.
    extended_params : ExtendedParams, optional
        If necessary other parameters.
    session : Session, optional
        session=None - means default session would be used
    on_response : Callable, optional
        Callable object to process retrieved data

    Returns
    -------
        CustomInstrumentBasket

    Examples
    --------
    >>> from refinitiv.data.content.custom_instruments.manage import create_udc
    >>> import refinitiv.data.content.custom_instruments as ci
    >>> import datetime
    ...
    >>> response_1 = create_udc(
    ...     symbol="MyUDCInstrument_VB",
    ...     instrument_name="Co Systems Inc",
    ...     udc=ci.manage.UDC(
    ...         root="CC",
    ...         months=ci.manage.Months(
    ...             number_of_years=3,
    ...             include_all_months=True,
    ...             start_month=1,
    ...         ),
    ...         rollover=ci.manage.VolumeBasedRollover(
    ...             method=ci.VolumeBasedRolloverMethod.VOLUME,
    ...             number_of_days=1,
    ...             join_at_day=1,
    ...             roll_occurs_within_months=4,
    ...             roll_on_expiry=True,
    ...         ),
    ...         spread_adjustment=ci.manage.SpreadAdjustment(
    ...             adjustment="arithmetic",
    ...             method=ci.SpreadAdjustmentMethod.CLOSE_TO_CLOSE,
    ...             backwards=True,
    ...         ),
    ...     ),
    ...
    >>> response_2 = create_udc(
    ...     symbol="MyUDCInstrument_DB",
    ...     instrument_name="ELBD Gbmx",
    ...     udc=ci.manage.UDC(
    ...         root="CC",
    ...         months=ci.manage.Months(
    ...             number_of_years=3,
    ...             include_all_months=True,
    ...             start_month=1,
    ...         ),
    ...         rollover=ci.manage.DayBasedRollover(
    ...             method=ci.DayBasedRolloverMethod.DAYS_BEFORE_END_OF_MONTH,
    ...             number_of_days=3,
    ...             months_prior=1,
    ...         ),
    ...         spread_adjustment=ci.manage.SpreadAdjustment(
    ...             adjustment="arithmetic",
    ...             method=ci.SpreadAdjustmentMethod.CLOSE_TO_CLOSE,
    ...             backwards=True,
    ...         ),
    ...     ),
    ...
    >>> response_3 = create_udc(
    ...     symbol="MyUDCInstrument_Manual",
    ...     instrument_name="REPKO Sys",
    ...     udc=ci.manage.UDC(
    ...         root="CC",
    ...         rollover=ci.manage.ManualRollover(
    ...             ci.manage.ManualItem(month=7, year=2022, start_date="2022-02-01"),
    ...             ci.manage.ManualItem(month=7, year=2021, start_date=datetime.date(2021, 3, 1)),
    ...             ci.manage.ManualItem(month=3, year=2020, start_date=datetime.timedelta(days=-950))
    ...         ),
    ...         spread_adjustment=ci.manage.SpreadAdjustment(
    ...             adjustment="arithmetic",
    ...             method=ci.SpreadAdjustmentMethod.CLOSE_TO_CLOSE,
    ...             backwards=True,
    ...         ),
    ...     ),
    """
    data = _create(
        symbol=symbol,
        type_=CustomInstrumentTypes.UDC,
        udc=udc,
        currency=currency,
        instrument_name=instrument_name,
        exchange_name=exchange_name,
        holidays=holidays,
        time_zone=time_zone,
        description=description,
        extended_params=extended_params,
        session=session,
        on_response=on_response,
    )
    return CustomInstrumentUDC(data, session=session)



from typing import TYPE_CHECKING

from ._data_provider import ImageData
from ...._content_type import ContentType
from ...._types import ExtendedParams
from ....delivery._data._data_provider_layer import DataProviderLayer
from ....delivery._data._response import BaseResponse

if TYPE_CHECKING:
    from typing import Optional


class Definition(DataProviderLayer[BaseResponse[ImageData]]):
    """
    This class describes parameters to retrieve data for news analyze.

    Parameters
    ----------
    image_id: str
        Image ID

    width: int, optional
        New width of the image

    height: int, optional
        New height of the image

    extended_params: dict, optional
        Other parameters can be provided if necessary

     Examples
    --------
    >>> from refinitiv.data.content import news
    >>> definition = news.images.Definition("image_id")
    >>> response = definition.get_data()
    >>> image = response.data.image
    >>> image.save("images")
    """

    def __init__(
        self,
        image_id: str,
        width: "Optional[int]" = None,
        height: "Optional[int]" = None,
        extended_params: "ExtendedParams" = None,
    ):
        super().__init__(
            data_type=ContentType.NEWS_IMAGES,
            image_id=image_id,
            width=width,
            height=height,
            extended_params=extended_params,
        )



from typing import Optional

from .._data._data_provider_layer import DataProviderLayer
from .._data._data_type import DataType
from ..._tools import validate_types


class Definition(DataProviderLayer):
    """
    Definition object.

    Parameters
    __________
        package_name : str, optional
            Filter results by package name. (Partial match).
         package_id: str, optional
            CFS package with package id.
        package_type : str, optional
            Filter results by package type.
            Available values : core, bulk, private.
        bucket_name : str, optional
            List the package that associated with this bucket.
        page : int, optional
            The offest number determine which set of result should be
            returned from all results.
        included_total_result : bool, optional
            The flag to indicate if total record count should be returned.
            By default False.
        skip_token : str, optional
            Skip token is only used if a previous operation returned a partial result.
            If a previous response contains a nextLink element, the value of the nextLink
            element will include a skip token parameter that specifies a starting point to use
            for subsequent calls.
        page_size : int, optional
            The number of package that will be returned into a single response.
            By default 25.
        included_entitlement_result : bool, optional
            The flag to indicate the entitlement checking on each package.
            By default False.

    Methods
    -------
    get_data(session=session)
        Returns a response to the data platform
    get_data_async(session=None)
        Returns a response asynchronously to the data platform

    Examples
    --------
     >>> from refinitiv.data.delivery import cfs
     >>> definition = cfs.packages.Definition()
     >>> packages = definition.get_data()

     Using get_data_async
     >>> import asyncio
     >>> task = definition.get_data_async()
     >>> response = asyncio.run(task)
    """

    def __init__(
        self,
        package_name: Optional[str] = None,
        package_id: Optional[str] = None,
        package_type: Optional[str] = None,
        bucket_name: Optional[str] = None,
        page: Optional[int] = None,
        included_total_result: bool = False,
        skip_token: Optional[str] = None,
        page_size: int = 25,
        included_entitlement_result: bool = False,
    ):
        validate_types(page_size, [int, type(None)], "page_size")
        super().__init__(
            data_type=DataType.CFS_PACKAGES,
            package_name=package_name,
            _package_id=package_id,
            package_type=package_type,
            bucket_name=bucket_name,
            page=page,
            included_total_result=included_total_result,
            skip_token=skip_token,
            page_size=page_size,
            included_entitlement_result=included_entitlement_result,
        )



from enum import Enum, unique


@unique
class Package(Enum):
    """
    BASIC - A limited set of fields with a single historical point which could typically be considered 'Free to Air' content

    STANDARD - The common fields for a content set with a limited amount of history

    PROFESSIONAL - All fields and all history for a particular content set
    """

    BASIC = "basic"
    STANDARD = "standard"
    PROFESSIONAL = "professional"



# coding: utf8


from enum import Enum, unique


@unique
class ForwardComputeMethod(Enum):
    USE_DIVIDENDS_ONLY = "UseDividendsOnly"
    USE_FORWARD_CRV = "UseForwardCrv"
    USE_FORWARD_CRV_AND_DIVIDENDS = "UseForwardCrvAndDividends"



__all__ = (
    "BusinessDayConvention",
    "DateRollingConvention",
    "DayCountBasis",
    "Definition",
    "PriceSide",
    "PricingParameters",
)

from ._definition import Definition
from ._enums import (
    BusinessDayConvention,
    DateRollingConvention,
    DayCountBasis,
    PriceSide,
)
from ._term_deposit_pricing_parameters import PricingParameters



import traceback
from typing import Generic, Optional, Callable, TYPE_CHECKING, Any

from ._response import BaseResponse, Response, T
from ..._configure import _RDPConfig
from ..._core.session._default_session_manager import get_valid_session
from ..._tools import DEBUG
from ...errors import RDError
from ...usage_collection._filter_types import FilterType
from ...usage_collection._logger import get_usage_logger
from ...usage_collection._utils import ModuleName

if TYPE_CHECKING:
    from ..._core.session import Session


def _check_response(
    response: BaseResponse, config: "_RDPConfig", response_class=Response
) -> None:
    if isinstance(response, response_class):
        is_raise_exception = config.get_param("raise_exception_on_error")
        if not response.is_success and is_raise_exception:
            error_code = response.errors[0].code
            error_message = response.errors[0].message
            exception_class = getattr(response, "exception_class", None)

            if exception_class:
                error = exception_class(error_code, error_message)

            else:
                error = RDError(error_code, error_message)

            error.response = response
            raise error


def get_data(data_type, provider, session, **kwargs):
    from ._data_provider_factory import get_url, get_api_config

    session = get_valid_session(session)
    config = session.config
    url = get_url(data_type, config)
    api_config = get_api_config(data_type, config)
    auto_retry = api_config.get("auto-retry", False)
    # Library usage logging
    get_usage_logger().log_func(
        name=f"{ModuleName.DELIVERY}.{DataProviderLayer._USAGE_CLS_NAME}.get_data",
        func_path=f"{DataProviderLayer.__module__}.{DataProviderLayer.__qualname__}.get_data",
        kwargs={"url": url, "auto_retry": auto_retry, **kwargs},
        desc={FilterType.SYNC, FilterType.LAYER_DELIVERY},
    )
    response = provider.get_data(session, url, auto_retry=auto_retry, **kwargs)
    return response


async def get_data_async(data_type, provider, session, **kwargs):
    from ._data_provider_factory import get_url, get_api_config

    session = get_valid_session(session)
    config = session.config
    url = get_url(data_type, config)
    api_config = get_api_config(data_type, config)
    auto_retry = api_config.get("auto-retry", False)
    # Library usage logging
    get_usage_logger().log_func(
        name=f"{ModuleName.DELIVERY}.{DataProviderLayer._USAGE_CLS_NAME}.get_data_async",
        func_path=f"{DataProviderLayer.__module__}.{DataProviderLayer.__qualname__}.get_data_async",
        kwargs={"url": url, "auto_retry": auto_retry, **kwargs},
        desc={FilterType.ASYNC, FilterType.LAYER_DELIVERY},
    )
    response = await provider.get_data_async(
        session, url, auto_retry=auto_retry, **kwargs
    )
    return response


def get_data_by_data_type(data_type, session, **kwargs):
    from refinitiv.data.delivery._data._data_provider_factory import make_provider

    provider = make_provider(data_type)
    return get_data(data_type, provider, session, **kwargs)


def emit_event(handler, *args, **kwargs):
    session = args[2]
    try:
        handler(*args, **kwargs)
    except Exception as e:
        session.error(f"{handler} callback raised exception: {e!r}")
        session.debug(traceback.format_exc())

        if DEBUG:
            raise e


class DataProviderLayer(Generic[T]):
    # Should not change even if class name is changed
    _USAGE_CLS_NAME = "DataProviderLayer"

    def __init__(self, data_type, **kwargs):
        self._initialize(data_type, **kwargs)
        self._log__init__usage(
            data_type=data_type,
            **kwargs,
        )

    @staticmethod
    def _log__init__usage(*args, **kwargs):
        get_usage_logger().log_func(
            name=f"{ModuleName.DELIVERY}.{DataProviderLayer._USAGE_CLS_NAME}.__init__",
            func_path=f"{DataProviderLayer.__module__}."
            f"{DataProviderLayer.__qualname__}."
            f"__init__",
            args=args,
            kwargs=kwargs,
            desc={FilterType.INIT, FilterType.LAYER_DELIVERY},
        )

    def _initialize(self, data_type, **kwargs):
        from ._data_provider_factory import make_provider

        self._kwargs = kwargs
        self._kwargs["__data_type__"] = data_type
        self._kwargs["__content_type__"] = data_type

        self._data_type = data_type
        self._content_type = data_type
        self._provider = make_provider(data_type)

    def _check_response(self, response: BaseResponse, config):
        _check_response(response, config)

    def get_data(
        self,
        session: Optional["Session"] = None,
        on_response: Optional[Callable] = None,
    ) -> T:
        session = get_valid_session(session)
        response = get_data(self._data_type, self._provider, session, **self._kwargs)
        on_response and emit_event(on_response, response, self, session)
        self._check_response(response, session.config)
        return response

    async def get_data_async(
        self,
        session: Optional["Session"] = None,
        on_response: Optional[Callable] = None,
        closure: Optional[Any] = None,
    ) -> T:
        if not self._kwargs.get("closure") and closure:
            self._kwargs["closure"] = closure
        session = get_valid_session(session)
        response = await get_data_async(
            self._data_type, self._provider, session, **self._kwargs
        )
        on_response and emit_event(on_response, response, self, session)
        return response

    def __repr__(self):
        s = super().__repr__()
        s = s.replace(">", f" {{name='{self._kwargs.get('universe')}'}}>")
        return s



import json
import typing

import numpy
import pandas as pd

from .._core.session._default_session_manager import _eikon_default_session_manager
from .._core.session.tools import is_closed

if typing.TYPE_CHECKING:
    from .._core.session._session import Session


def is_string_type(value):
    try:
        return isinstance(value, basestring)
    except NameError:
        return isinstance(value, str)


def get_json_value(json_data, name):
    if name in json_data:
        return json_data[name]
    else:
        return None


def is_list_of_string(values):
    return all(is_string_type(value) for value in values)


def check_for_string(parameter, name):
    if not is_string_type(parameter):
        raise ValueError(
            "The parameter {} should be a string, found {}".format(name, str(parameter))
        )


def check_for_string_or_list_of_strings(parameter, name):
    if type(parameter) != list and (not parameter or not is_string_type(parameter)):
        raise ValueError(
            "The parameter {} should be a string or a list of string, found {}".format(
                name, type(parameter)
            )
        )
    if type(parameter) == list and not is_list_of_string(parameter):
        raise ValueError(
            "All items in the parameter {} should be of data type string, found {}".format(
                name, [type(v) for v in parameter]
            )
        )


def check_for_int(parameter, name):
    if type(parameter) is not int:
        raise ValueError(
            "The parameter {} should be an int, found {} type value ({})".format(
                name, type(parameter), str(parameter)
            )
        )


def build_list_with_params(values, name):
    if values is None:
        raise ValueError(name + " is None, it must be a string or a list of strings")

    if is_string_type(values):
        return [(v, None) for v in values.split()]
    elif type(values) is list:
        try:
            return [
                (value, None) if is_string_type(value) else (value[0], value[1])
                for value in values
            ]
        except Exception:
            raise ValueError(
                name
                + " must be a string or a list of strings or a tuple or a list of tuple"
            )
    else:
        try:
            return values[0], values[1]
        except Exception:
            raise ValueError(
                name
                + " must be a string or a list of strings or a tuple or a list of tuple"
            )


def build_list(values, name):
    if values is None:
        raise ValueError(name + " is None, it must be a string or a list of strings")

    if is_string_type(values):
        return [values.strip()]
    elif type(values) is list:
        if all(is_string_type(value) for value in values):
            return [value for value in values]
        else:
            raise ValueError(name + " must be a string or a list of strings")
    else:
        raise ValueError(name + " must be a string or a list of strings")


def build_dictionary(dic, name):
    if dic is None:
        raise ValueError(
            name + " is None, it must be a string or a dictionary of strings"
        )

    if is_string_type(dic):
        return json.loads(dic)
    elif type(dic) is dict:
        return dic
    else:
        raise ValueError(name + " must be a string or a dictionary")


def tz_replacer(s: str) -> str:
    if isinstance(s, str):
        if s.endswith("Z"):
            s = s[:-1]
        elif s.endswith("-0000"):
            s = s[:-5]
        if s.endswith(".000"):
            s = s[:-4]
    return s


def set_default_session(session: "Session"):
    _eikon_default_session_manager.set_default_session(session)


def get_default_session(app_key=None) -> "Session":
    return _eikon_default_session_manager.get_default_session(app_key)


def close_session():
    _eikon_default_session_manager.get_default_session().close()


def set_app_key(app_key):
    _session = get_default_session(app_key)
    if is_closed(_session):
        _session.open()


def set_log_level(log_level):
    default_session = _eikon_default_session_manager.get_default_session()
    default_session.set_log_level(log_level)


def convert_content_data_to_df_udf(raw: dict) -> pd.DataFrame:
    selected_fields = ["versionCreated", "text", "storyId", "sourceCode"]

    raw_headlines = raw.get("headlines", [])
    first_created = [
        tz_replacer(headline["firstCreated"]) for headline in raw_headlines
    ]
    headlines = [
        [headline[field] for field in selected_fields] for headline in raw_headlines
    ]
    if len(headlines):
        df = pd.DataFrame(
            headlines,
            numpy.array(first_created, dtype="datetime64[ns]"),
            selected_fields,
        )

        if not df.empty:
            df = df.convert_dtypes()

    else:
        df = pd.DataFrame(
            [], numpy.array(first_created, dtype="datetime64[ns]"), selected_fields
        )

    df["versionCreated"] = df.versionCreated.apply(pd.to_datetime)
    df.fillna(pd.NA, inplace=True)

    return df



from refinitiv.data.content.ipa._curves._cross_currency_curves._curves import (
    FxForwardCurveParameters,
)
from refinitiv.data.content.ipa._enums._extrapolation_mode import ExtrapolationMode
from refinitiv.data.content.ipa._enums._interpolation_mode import InterpolationMode
from refinitiv.data.content.ipa._curves._cross_currency_curves._curves import (
    ValuationTime,
)


def test_fx_forward_curve_parameters():
    # given
    expected_dict = {
        "extrapolationMode": "Constant",
        "ignoreInvalidInstrument": True,
        "ignorePivotCurrencyHolidays": True,
        "interpolationMode": "CubicDiscount",
        "useDelayedDataIfDenied": True,
        "valuationDate": "string",
        "valuationDateTime": "string",
        "valuationTime": {},
    }

    # when
    testing_obj = FxForwardCurveParameters(
        extrapolation_mode=ExtrapolationMode.CONSTANT,
        interpolation_mode=InterpolationMode.CUBIC_DISCOUNT,
        valuation_time=ValuationTime(),
        ignore_invalid_instrument=True,
        ignore_pivot_currency_holidays=True,
        use_delayed_data_if_denied=True,
        valuation_date="string",
        valuation_date_time="string",
    )

    # then
    assert testing_obj.get_dict() == expected_dict

    assert testing_obj.extrapolation_mode == ExtrapolationMode.CONSTANT
    assert testing_obj.interpolation_mode == InterpolationMode.CUBIC_DISCOUNT
    assert testing_obj.valuation_time == ValuationTime()
    assert testing_obj.ignore_invalid_instrument == True
    assert testing_obj.ignore_pivot_currency_holidays == True
    assert testing_obj.use_delayed_data_if_denied == True
    assert testing_obj.valuation_date == "string"
    assert testing_obj.valuation_date_time == "string"



import json
import os.path
import shutil

from refinitiv.data.content import news
from tests.unit.conftest import StubResponse, StubSession

path_to_save = "tests/unit/content/news/images/files"

with open("tests/unit/content/news/images/images.json") as f:
    images_response = json.loads(f.read())


def make_check_url_in_request(expected_url: str):
    def check_url_in_http_request(response):
        url = response.url
        assert url == expected_url
        response = StubResponse(images_response)
        response.content = b""
        return response

    return check_url_in_http_request


def test_news_images():
    session = StubSession(is_open=True, response=StubResponse(images_response))

    response = news.images.Definition("id").get_data(session)
    image = response.data.image

    assert response.data.raw == images_response
    assert image.provider == ["NOT AVAILABLE"]
    assert image.body_type == ["image/jpeg"]
    assert image.source == ["RTRS"]
    assert image.version_created == ["20220927T132732.524+0000"]
    assert image.first_created == ["20220927T132732.524+0000"]
    assert image.filename == "2022-09-27T132732Z_1_CN0_RTRLXPP_2_LYNXPACKAGER.JPG"
    assert image.available_rsf == ["No"]
    assert image.size == 33603
    assert image.show()


def test_save_with_size():
    image_id = "22T112537Z_1_MQ4_RTRLXPP_2_LYNXPACKAGER__JPG"
    with open(f"tests/unit/content/news/images/{image_id}", "rb") as f:
        image_data = f.read()
    stub_response = StubResponse()
    stub_response.content = image_data
    stub_response.http_headers["content-type"] = "image/jpeg"
    session = StubSession(is_open=True, response=stub_response)
    response = news.images.Definition("id", width=200).get_data(session)
    image = response.data.image
    image.save(path_to_save)

    assert os.path.exists(f"{path_to_save}/{image.filename}")

    shutil.rmtree(path_to_save)


def test_save():
    session = StubSession(is_open=True, response=StubResponse(images_response))

    response = news.images.Definition("id").get_data(session)
    image = response.data.image
    image.save(path_to_save)

    assert os.path.exists(f"{path_to_save}/{image.filename}")

    shutil.rmtree(path_to_save)


def test_url():
    session = StubSession(is_open=True)
    session.http_request = make_check_url_in_request(
        "test_get_rdp_url_root/data/news/v1/images/id?width=100&height=200"
    )

    news.images.Definition("id", width=100, height=200).get_data(session)


def test_extended_params():
    session = StubSession(is_open=True)
    session.http_request = make_check_url_in_request(
        "test_get_rdp_url_root/data/news/v1/images/id?test_param=test_value"
    )

    news.images.Definition("id", extended_params={"test_param": "test_value"}).get_data(
        session
    )



from typing import Union, Any

import numpy as np
import pandas as pd
from numpy import iterable

from ._enums import Axis
from ._object_definition import ObjectDefinition
from ...delivery._data._data_provider import RequestFactory
from ...delivery.endpoint_request import RequestMethod
from ..._tools._common import ArgsParser, merge_dict_to_dict

types_by_axis = {
    Axis.DATE: "datetime64",
    "startDate": "datetime64",
    "endDate": "datetime64",
    Axis.DELTA: float,
    Axis.EXPIRY: float,
    Axis.MONEYNESS: float,
    Axis.STRIKE: float,
    Axis.TENOR: float,
    "discountFactor": float,
    "ratePercent": float,
}


def get_type_by_axis(axis):
    axis_values_type = types_by_axis.get(axis)

    if not axis_values_type:
        raise ValueError(f"Cannot find axis's values type for axis {axis}.")

    return axis_values_type


def parse_value(value: Any) -> Union[float, int, np.datetime64]:
    if isinstance(value, str):
        try:
            value = int(value)
        except ValueError:
            pass
        else:
            return value

        try:
            value = pd.to_datetime(value).to_numpy()
        except ValueError:
            try:
                value = pd.to_datetime(value, dayfirst=True).to_numpy()
            except ValueError:
                pass
        else:
            return value

        try:
            value = float(value)
        except ValueError:
            raise ValueError(f"not valid format: {value}")

    return value


value_arg_parser = ArgsParser(parse_value)
x_arg_parser = value_arg_parser
y_arg_parser = value_arg_parser


# ---------------------------------------------------------------------------
#   RequestFactory
# ---------------------------------------------------------------------------


def parse_universe(universe):
    retval = []

    if not iterable(universe):
        universe = [universe]

    # convert universe's objects into json
    for i, item in enumerate(universe):
        extended_params = None
        if not hasattr(item, "get_dict"):
            kwargs = item._kwargs
            item = kwargs.get("universe")
            extended_params = kwargs.get("extended_params")
        item_dict = item.get_dict()
        if extended_params:
            item_dict.update(extended_params)
        retval.append(item_dict)

    return retval


def parse_outputs(outputs):
    retval = []

    if not isinstance(outputs, list):
        outputs = [outputs]

    for item in outputs:
        if hasattr(item, "value"):
            item = item.value

        retval.append(item)

    return retval


universe_arg_parser = ArgsParser(parse_universe)
outputs_arg_parser = ArgsParser(parse_outputs)


class CurvesAndSurfacesRequestFactory(RequestFactory):
    def extend_body_parameters(self, body_parameters, extended_params=None, **kwargs):
        if not extended_params:
            return body_parameters

        if kwargs.get("__plural__") is True:
            body_parameters.update(extended_params)
            return body_parameters

        universes = body_parameters.get("universe", [{}])
        universes[0] = merge_dict_to_dict(universes[0], extended_params)
        return body_parameters

    def get_request_method(self, *, method=None, **kwargs):
        return method or RequestMethod.POST

    def get_body_parameters(self, *args, universe=None, outputs=None, **kwargs):
        body_parameters = {}

        # universe
        universe = universe_arg_parser.get_list(universe)
        body_parameters["universe"] = universe

        # outputs
        if outputs:
            outputs = outputs_arg_parser.get_list(outputs)
            body_parameters["outputs"] = outputs

        return body_parameters


class CrossCurrencyCurvesDefinitionsRequestFactory(RequestFactory):
    def get_request_method(self, *, method=None, **kwargs):
        return method or RequestMethod.POST

    def get_body_parameters(self, *args, request_items: ObjectDefinition, **kwargs):
        if isinstance(request_items, ObjectDefinition):
            result = request_items.get_dict()
            return result
        return {}

    def extend_body_parameters(self, body_parameters, extended_params=None, **kwargs):
        if extended_params:
            result = dict(body_parameters)
            result.update(extended_params)
            return result
        return body_parameters


class DatesAndCalendarsRequestFactory(RequestFactory):
    def get_request_method(self, *, method=None, **kwargs):
        return method or RequestMethod.POST

    def get_body_parameters(self, *args, universe, **kwargs):
        body_parameters = []
        for request_item in universe:
            body_parameters.append(request_item.get_dict())
        return body_parameters

    def extend_body_parameters(self, body_parameters, extended_params=None, **kwargs):
        if extended_params:
            if isinstance(extended_params, list):
                for idx, extended_param_item in enumerate(extended_params):
                    if extended_param_item:
                        body_parameters[idx].update(extended_param_item)
            else:
                for item in body_parameters:
                    item.update(extended_params)
        return body_parameters


class DateScheduleRequestFactory(RequestFactory):
    def get_request_method(self, *, method=None, **kwargs):
        return method or RequestMethod.POST

    def get_body_parameters(self, *args, universe, **kwargs):
        return universe.get_dict()

    def extend_body_parameters(self, body_parameters, extended_params=None, **kwargs):
        if extended_params:
            body_parameters.update(extended_params)
        return body_parameters



# coding: utf8


from typing import Optional

from .._enums import PremiumSettlementType
from .._object_definition import ObjectDefinition


class InputFlow(ObjectDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    amount : float, optional

    premium_settlement_type : PremiumSettlementType, optional
        The cash settlement type of the option premium -spot -forward
    currency : str, optional

    date : str, optional

    """

    def __init__(
        self,
        amount: Optional[float] = None,
        premium_settlement_type: Optional[PremiumSettlementType] = None,
        currency: Optional[str] = None,
        date: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.amount = amount
        self.premium_settlement_type = premium_settlement_type
        self.currency = currency
        self.date = date

    @property
    def premium_settlement_type(self):
        """
        The cash settlement type of the option premium -spot -forward
        :return: enum PremiumSettlementType
        """
        return self._get_enum_parameter(PremiumSettlementType, "premiumSettlementType")

    @premium_settlement_type.setter
    def premium_settlement_type(self, value):
        self._set_enum_parameter(PremiumSettlementType, "premiumSettlementType", value)

    @property
    def amount(self):
        """
        :return: float
        """
        return self._get_parameter("amount")

    @amount.setter
    def amount(self, value):
        self._set_parameter("amount", value)

    @property
    def currency(self):
        """
        :return: str
        """
        return self._get_parameter("currency")

    @currency.setter
    def currency(self, value):
        self._set_parameter("currency", value)

    @property
    def date(self):
        """
        :return: str
        """
        return self._get_parameter("date")

    @date.setter
    def date(self, value):
        self._set_parameter("date", value)



import logging

from tests.integration.constants_list import ISO_DATE_FORMAT
from tests.integration.delivery.cfs.conftest import (
    convert_string_to_date_format,
)


def check_buckets_in_response_created_after_date(expected_date, response):
    date = convert_string_to_date_format(expected_date, ISO_DATE_FORMAT)
    buckets = response.data.raw["value"]
    for bucket in buckets:
        bucket_created_date = convert_string_to_date_format(
            bucket["created"], ISO_DATE_FORMAT
        )
        assert (
            bucket_created_date > date
        ), f"found bucket created before the required date {date}: \n {bucket}"


def check_buckets_in_response_modified_after_date(expected_date, response):
    date = convert_string_to_date_format(expected_date, ISO_DATE_FORMAT)
    buckets = response.data.raw["value"]
    for bucket in buckets:
        bucket_modified_date = convert_string_to_date_format(
            bucket["modified"], ISO_DATE_FORMAT
        )
        assert (
            bucket_modified_date > date
        ), f"found bucket modified before the required date {date}: \n {bucket}"


def check_buckets_in_response_available_before_date(expected_date, response):
    date = convert_string_to_date_format(expected_date, ISO_DATE_FORMAT)
    buckets = response.data.raw["value"]
    for bucket in buckets:
        try:
            bucket_available_from_date = convert_string_to_date_format(
                bucket["availableFrom"], ISO_DATE_FORMAT
            )
            assert (
                bucket_available_from_date <= date
            ), f"found bucket available only after the required date {date}: \n {bucket}"
        except KeyError:
            logging.warning(f"The bucket does not have 'availableFrom' field: {bucket}")


def check_buckets_in_response_available_after_date(expected_date, response):
    date = convert_string_to_date_format(expected_date, ISO_DATE_FORMAT)
    buckets = response.data.raw["value"]
    for bucket in buckets:
        try:
            bucket_available_to_date = convert_string_to_date_format(
                bucket["availableTo"], ISO_DATE_FORMAT
            )
            assert (
                bucket_available_to_date >= date
            ), f"found bucket available before  the required date {date}: \n {bucket}"
        except KeyError:
            logging.warning(f"The bucket does not have 'availableFrom' field: {bucket}")


def check_all_attributes_found_in_bucket(bucket, expected_attributes_list):
    if isinstance(expected_attributes_list, str):
        expected_attributes_list = [expected_attributes_list]
    for attribute in expected_attributes_list:
        assert (
            attribute in map(str.lower, bucket["attributes"])
        ), f"Found bucket which does not contain expected attribute {attribute}: {bucket} "


def check_all_buckets_contain_attributes(response, expected_attributes_list):
    for bucket in response.data.raw["value"]:
        check_all_attributes_found_in_bucket(bucket, expected_attributes_list)


def check_buckets_in_two_responses_are_not_the_same(response1, response2):
    for bucket in response1.data.raw["value"]:
        assert (
            bucket not in response2.data.raw["value"]
        ), f"Bucket from first request found in list of buckets in the second request: \n {bucket}"



# coding: utf8

from typing import Optional

from ._enums import PriceSide, FxSwapCalculationMethod, ImpliedDepositDateConvention
from ..._object_definition import ObjectDefinition


class PricingParameters(ObjectDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    fx_swap_calculation_method : FxSwapCalculationMethod, optional
        The method used to calculate an outright price or deposit rates. the possible
        values are:
        - fxswap: outrights are computed using swap points,
        - fxswapimpliedfromdeposit: implied fx swap points are computed from deposit
          rates.  optional. defaults to 'fxswap'.
    implied_deposit_date_convention : ImpliedDepositDateConvention, optional
        An indicator for the "depositccy1impliedfromfxswap",
        "depositccy2impliedfromfxswap" calculation methods, used to check whether dts or
        dtm period is selected for the implied deposits calculation. by default, for the
        implied deposits calculation we use the dts(day to spot) period, while for the
        deposits the dtm(day to money market) period is used. the possible values are:
        - fxmarketconvention:    dts case, traditional method getperiodsfromtenor is
          used with the parameter "from:fxtrade"
        - moneymarketconvention: dtm case, getperiodsfromtenor with the parameter
          "from:mmtrade" is used optional.
    price_side : PriceSide, optional
        The quoted price side of the instrument. the possible values are:
        - bid,
        - ask,
        - mid. optional. defaults to 'mid'.
    adjust_all_deposit_points_to_cross_calendars : bool, optional
        An indicator if depositccy1marketdata and depositccy2marketdata are adjusted to
        the cross calendar dates. the possible values are:
        - true: the market data is adjusted according to the cross calendar dates,
        - false: unmodified market data is returned, cross calendar is ignored.
          optional. defaults to 'false'.
    adjust_all_swap_points_to_cross_calendars : bool, optional
        An indicator if fxswapsccy1 and fxswapsccy2 are adjusted to the cross calendar
        dates. the possible values are:
        - true: the market data is adjusted according to the cross calendar dates,
        - false: unmodified market data is returned, cross calendar is ignored.
          optional. defaults to 'false'.
    ignore_ref_ccy_holidays : bool, optional
        An indicator if holidays of the reference currency are included or not in the
        pricing when dates are computed. by default, the reference currency is usd. the
        possible values are:
        - true: holidays are ignored,
        - false: holidays are included in pricing. optional. defaults to 'false'.
    market_data_date : str, optional
        The date at which the market data is retrieved. the value is expressed in iso
        8601 format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g., '2021-01-01t00:00:00z'). it
        should be less or equal tovaluationdate). optional. by
        default,marketdatadateisvaluationdateor today.
    report_ccy : str, optional
        The reporting currency code, expressed in iso 4217 alphabetical format (e.g.,
        'usd'). it is set for the fields ending with 'xxxinreportccy'. optional. the
        default value is the notional currency.
    use_direct_quote : bool, optional
        An indicator if the spot and the swap points should be retrieved without a pivot
        currency
        - true: the market data is retrieved directly
        - false: unmodified market data is returned. optional. defaults to 'false'.
    valuation_date : str, optional
        The valuation date for pricing.  Optional. If not set the valuation date is
        equal to market_data_date or Today. For assets that contains a
        settlementConvention, the default valuation date  is equal to the settlementdate
        of the Asset that is usually the TradeDate+SettlementConvention.
    """

    def __init__(
        self,
        fx_swap_calculation_method: Optional[FxSwapCalculationMethod] = None,
        implied_deposit_date_convention: Optional[ImpliedDepositDateConvention] = None,
        price_side: Optional[PriceSide] = None,
        adjust_all_deposit_points_to_cross_calendars: Optional[bool] = None,
        adjust_all_swap_points_to_cross_calendars: Optional[bool] = None,
        ignore_ref_ccy_holidays: Optional[bool] = None,
        market_data_date: Optional[str] = None,
        report_ccy: Optional[str] = None,
        use_direct_quote: Optional[bool] = None,
        valuation_date: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.fx_swap_calculation_method = fx_swap_calculation_method
        self.implied_deposit_date_convention = implied_deposit_date_convention
        self.price_side = price_side
        self.adjust_all_deposit_points_to_cross_calendars = (
            adjust_all_deposit_points_to_cross_calendars
        )
        self.adjust_all_swap_points_to_cross_calendars = (
            adjust_all_swap_points_to_cross_calendars
        )
        self.ignore_ref_ccy_holidays = ignore_ref_ccy_holidays
        self.market_data_date = market_data_date
        self.report_ccy = report_ccy
        self.use_direct_quote = use_direct_quote
        self.valuation_date = valuation_date

    @property
    def fx_swap_calculation_method(self):
        """
        The method we chose to price outright using or not implied deposits. Possible
        values are:   FxSwap (compute outright using swap points),
        DepositCcy1ImpliedFromFxSwap (compute currency1 deposits using swap points),
        DepositCcy2ImpliedFromFxSwap (compute currency2 deposits using swap points).
        Optional. Defaults to 'FxSwap'.
        :return: enum FxSwapCalculationMethod
        """
        return self._get_enum_parameter(
            FxSwapCalculationMethod, "fxSwapCalculationMethod"
        )

    @fx_swap_calculation_method.setter
    def fx_swap_calculation_method(self, value):
        self._set_enum_parameter(
            FxSwapCalculationMethod, "fxSwapCalculationMethod", value
        )

    @property
    def implied_deposit_date_convention(self):
        """
        :return: object FxPoint
        """
        return self._get_enum_parameter(
            ImpliedDepositDateConvention, "impliedDepositDateConvention"
        )

    @implied_deposit_date_convention.setter
    def implied_deposit_date_convention(self, value):
        self._set_enum_parameter(
            ImpliedDepositDateConvention, "impliedDepositDateConvention", value
        )

    @property
    def price_side(self):
        """
        The type of price returned for pricing Analysis:
        Bid(Bid value),
        Ask(Ask value),
        Mid(Mid value)
        Optional. Defaults to 'Mid'.
        :return: enum PriceSide
        """
        return self._get_enum_parameter(PriceSide, "priceSide")

    @price_side.setter
    def price_side(self, value):
        self._set_enum_parameter(PriceSide, "priceSide", value)

    @property
    def adjust_all_deposit_points_to_cross_calendars(self):
        """
        An indicator if depositccy1marketdata and depositccy2marketdata are adjusted to
        the cross calendar dates. the possible values are:
        - true: the market data is adjusted according to the cross calendar dates,
        - false: unmodified market data is returned, cross calendar is ignored.
          optional. defaults to 'false'.
        :return: bool
        """
        return self._get_parameter("adjustAllDepositPointsToCrossCalendars")

    @adjust_all_deposit_points_to_cross_calendars.setter
    def adjust_all_deposit_points_to_cross_calendars(self, value):
        self._set_parameter("adjustAllDepositPointsToCrossCalendars", value)

    @property
    def adjust_all_swap_points_to_cross_calendars(self):
        """
        This flag define if cross-calendar holidays should be taken into account for
        fx_swaps_ccy1 and fx_swaps_ccy2. it adjusts swap points according to cross
        holidays if it's set to true, by default set to false.
        :return: bool
        """
        return self._get_parameter("adjustAllSwapPointsToCrossCalendars")

    @adjust_all_swap_points_to_cross_calendars.setter
    def adjust_all_swap_points_to_cross_calendars(self, value):
        self._set_parameter("adjustAllSwapPointsToCrossCalendars", value)

    @property
    def ignore_ref_ccy_holidays(self):
        """
        The reference currency holidays flag : When dates are computed, its possible to
        choose if holidays of the reference currency are included or not in the pricing
        Optional. Defaults to 'false'.
        :return: bool
        """
        return self._get_parameter("ignoreRefCcyHolidays")

    @ignore_ref_ccy_holidays.setter
    def ignore_ref_ccy_holidays(self, value):
        self._set_parameter("ignoreRefCcyHolidays", value)

    @property
    def market_data_date(self):
        """
        The market data date for pricing. Optional. By default, the market_data_date
        date is the valuation_date or Today
        :return: str
        """
        return self._get_parameter("marketDataDate")

    @market_data_date.setter
    def market_data_date(self, value):
        self._set_parameter("marketDataDate", value)

    @property
    def report_ccy(self):
        """
        :return: str
        """
        return self._get_parameter("reportCcy")

    @report_ccy.setter
    def report_ccy(self, value):
        self._set_parameter("reportCcy", value)

    @property
    def use_direct_quote(self):
        """
        An indicator if the spot and the swap points should be retrieved without a pivot
        currency
        - true: the market data is retrieved directly
        - false: unmodified market data is returned. optional. defaults to 'false'.
        :return: bool
        """
        return self._get_parameter("useDirectQuote")

    @use_direct_quote.setter
    def use_direct_quote(self, value):
        self._set_parameter("useDirectQuote", value)

    @property
    def valuation_date(self):
        """
        The valuation date for pricing.  Optional. If not set the valuation date is
        equal to market_data_date or Today. For assets that contains a
        settlementConvention, the default valuation date  is equal to the settlementdate
        of the Asset that is usually the TradeDate+SettlementConvention.
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)



# coding: utf8

import abc

from ..._tools._common import is_all_same_type


class ObjectDefinition(abc.ABC):
    """
    This class is designed to represent the instrument definition templates for QPS request.
    """

    @classmethod
    def from_json(cls, json_params):
        o = ObjectDefinition()
        for key, value in json_params.items():
            o._set_parameter(key, value)

        return o

    def __init__(self):
        self._dict = {}

    def __eq__(self, other):
        if hasattr(other, "_kwargs"):
            other = getattr(other, "_kwargs").get("definition")

        return self._dict == other

    def get_dict(self):
        return self._dict

    ####################################################
    # Get parameter values
    ####################################################
    def _get_parameter(self, name):
        return self._dict.get(name, None)

    def _get_enum_parameter(self, enum_type, name):
        value = self._dict.get(name, None)
        return enum_type(value) if value is not None else None

    def _get_object_parameter(self, object_type, name):
        value = self._dict.get(name, None)
        return object_type.from_json(value) if value is not None else None

    def _set_list_of_enums(self, enum_type, name, values):
        values = values or []
        self._dict[name] = []

        for item in values:
            if item is None:
                self._dict.pop(name, False)

            elif isinstance(item, enum_type):
                self._dict[name].append(item.value)

            elif item in [v.value for v in enum_type]:
                self._dict[name].append(item)

            else:
                values = [v.value for v in enum_type]
                raise TypeError(f"{name} : {values}, must be in {enum_type}{values}")

    def _get_list_of_enums(self, enum_type, name):
        enum_value = self._dict.get(name, None)
        response = []
        for item in enum_value:
            if item is not None:
                response.append(enum_type(item))

        return response

    def _get_list_parameter(self, item_type, name):
        value = self._dict.get(name, None)
        return (
            [
                item_type.from_json(item) if hasattr(item_type, "from_json") else item
                for item in value
            ]
            if value is not None
            else None
        )

    ####################################################
    # Set parameter values
    ####################################################

    def _set_parameter(self, name, value):
        if value or isinstance(value, int):
            self._dict[name] = value

        elif self._dict.get(name):
            self._dict.pop(name)

    def _set_enum_parameter(self, enum_type, name, value):
        if value is None:
            self._dict.pop(name, False)
            return None

        result = None
        if isinstance(value, enum_type):
            result = value.value

        elif isinstance(value, str):
            value_upper = value.upper()
            upper_enum_values = {value.value.upper(): value for value in enum_type}
            if value_upper in upper_enum_values:
                result = upper_enum_values[value_upper].value

        if result:
            self._dict[name] = result
        else:
            values = [v.value for v in enum_type]
            raise TypeError(
                f"Parameter '{name}' of invalid type provided:'{type(value).__name__}',"
                f"expected: {values}"
            )

    def _set_object_parameter(self, object_type, name, value):
        if value is None:
            self._dict.pop(name, False)

        elif isinstance(value, object_type):
            self._dict[name] = value.get_dict()

        elif hasattr(value, "_kwargs"):
            attr = getattr(value, "_kwargs")
            self._dict[name] = attr.get("definition").get_dict()

        else:
            self._dict[name] = value

    def _set_list_parameter(self, item_type, name, value):
        if value is None:
            self._dict.pop(name, False)

        elif isinstance(value, list):
            if is_all_same_type(item_type, value):
                self._dict[name] = [
                    item.get_dict() if hasattr(item, "get_dict") else item
                    for item in value
                ]

            else:
                raise TypeError(f"Not all values are type of {item_type}")

        else:
            raise TypeError(f"{name} value must be a list of {item_type}")



import pandas as pd
import pytest

from refinitiv.data.content.search import Views
from refinitiv.data.content.search._data_provider import (
    SearchRequestFactory,
    LookupRequestFactory,
    BaseSearchRequestFactory,
    MetadataRequestFactory,
    _get_unique_keys,
    discovery_search_build_df,
    discovery_lookup_build_df,
)
from refinitiv.data.delivery._data._endpoint_data import RequestMethod


def tests_base_search_request_factory_get_request_method():
    # given
    factory = BaseSearchRequestFactory()

    # when
    testing_value = factory.get_request_method()

    # then
    assert testing_value == RequestMethod.POST


def tests_base_search_request_factory_get_body_parameters():
    # given
    factory = SearchRequestFactory()

    # when
    testing_value = factory.get_body_parameters()

    # then
    assert testing_value == {}


def tests_search_request_factory_get_body_parameters_pass_view():
    # given
    factory = SearchRequestFactory()

    # when
    testing_value = factory.get_body_parameters(**{"view": Views.SEARCH_ALL})

    # then
    assert "View" in testing_value
    assert testing_value["View"] == Views.SEARCH_ALL.value


def tests_search_request_factory_get_body_parameters_pass_args():
    # given
    factory = SearchRequestFactory()
    extended_body_params = [
        "Boost",
        "Features",
        "Filter",
        "GroupBy",
        "GroupCount",
        "Navigators",
        "OrderBy",
        "Query",
        "Scope",
        "Select",
        "Skip",
        "Terms",
        "Top",
    ]

    # when
    testing_value = factory.get_body_parameters(
        **{
            "boost": "value_boost",
            "features": "value_features",
            "filter": "value_filter",
            "group_by": "value_group_by",
            "group_count": "value_group_count",
            "navigators": "value_navigators",
            "order_by": "value_order_by",
            "query": "value_query",
            "scope": "value_scope",
            "select": "value_select",
            "skip": "value_skip",
            "terms": "value_terms",
            "top": "value_top",
        }
    )

    # then
    for param in extended_body_params:
        assert param in testing_value
        assert testing_value[param]


def tests_lookup_request_factory_get_body_parameters_pass_args():
    # given
    factory = LookupRequestFactory()
    extended_body_params = [
        "Boost",
        "Filter",
        "Scope",
        "Select",
        "Terms",
    ]

    # when
    testing_value = factory.get_body_parameters(
        **{
            "boost": "value_boost",
            "filter": "value_filter",
            "scope": "value_scope",
            "select": "value_select",
            "terms": "value_terms",
        }
    )

    # then
    for param in extended_body_params:
        assert param in testing_value
        assert testing_value[param]


def tests_metadata_request_factory_get_query_parameters_pass_view():
    # given
    factory = MetadataRequestFactory()

    # when
    testing_value = factory.get_query_parameters(**{"view": Views.SEARCH_ALL})

    # then
    assert testing_value == Views.SEARCH_ALL.value


def test_metadata_request_factory_add_query_parameters():
    # given
    factory = MetadataRequestFactory()

    # when
    testing_value = factory.add_query_parameters(url="url", query_parameters="test")

    # then
    assert testing_value == "url/test"


@pytest.mark.parametrize(
    "input_value, expected_value",
    [
        ([{}, {}], []),
        ([{"a": 1, "b": 2}, {"a": 11, "b": 22}], ["a", "b"]),
        (
            [{"a": 1, "b": 2}, {"a": 1, "b": 2, "c": 1}, {"a": 1, "b": 2}],
            ["a", "b", "c"],
        ),
        (
            [{"a": 1, "b": 2}, {"a": 1, "b": 2, "c": 1}, {"a": 1, "d": 4, "b": 2}],
            ["a", "b", "c", "d"],
        ),
    ],
)
def test__get_unique_keys(input_value, expected_value):
    # when
    result = _get_unique_keys(input_value)

    # then
    assert result == expected_value


def test_convert_search_json_to_pandas():
    data = {
        "Hits": [
            {
                "BusinessEntity": "ORGANISATION",
                "DocumentTitle": "International Business Machines Corp, "
                "Public Company",
                "PI": "37036",
                "RIC": None,
            },
            {
                "BusinessEntity": "ORGANISATION",
                "DocumentTitle": "Banco IBM SA, Private Company",
                "PI": "76208",
                "RIC": "",
            },
            {
                "BusinessEntity": "QUOTExEQUITY",
                "DocumentTitle": "International Business Machines Corp, Ordinary "
                "Share, NYSE Consolidated",
                "PermID": "55839165994",
                "PI": "1097326",
                "RIC": "IBM",
            },
        ]
    }
    df = discovery_search_build_df(data)
    for row in df.itertuples():
        if row.PI == "37036":
            assert row.RIC is pd.NA
        if row.PI == "76208":
            assert isinstance(row.RIC, str)
            assert row.RIC == ""
        if row.PI == "1097326":
            assert isinstance(row.RIC, str)
            assert row.RIC == "IBM"


@pytest.mark.parametrize(
    "name_field",
    [
        "Date",
        "IssueDate",
        "Issue Date",
        "DateIssue",
        "Date Issue",
        "IssDateue",
        "Iss Date ue",
    ],
)
def test_convert_search_json_to_pandas_date_data_successful(name_field):
    # given
    data = {"Hits": [{name_field: "2019-05-15T00:00:00.000Z"}]}

    # when
    df = discovery_search_build_df(data)

    # then
    assert df[name_field].dtypes == "datetime64[ns]"


@pytest.mark.parametrize(
    "name_field",
    [
        "date",
        "Da te",
        "Issuedate",
        "Issue date",
        "dateIssue",
        "date Issue",
        "Issdateue",
        "Iss date ue",
    ],
)
def test_convert_search_json_to_pandas_date_data_fail(name_field):
    # given
    data = {"Hits": [{name_field: "2019-05-15T00:00:00.000Z"}]}

    # when
    df = discovery_search_build_df(data)

    # then
    assert df[name_field].dtypes != "datetime64[ns, UTC]"


def test__convert_lookup_json_to_df():
    data = {
        "Matches": {
            "MSFT.O": {
                "BusinessEntity": "INSTRUMENTxEQUITY",
                "DocumentTitle": "Microsoft Corp, Ordinary Share",
                "CUSIP": "594918104",
            },
            "GOOG.O": {
                "BusinessEntity": "INSTRUMENTxEQUITY",
                "DocumentTitle": "Alphabet Inc, Ordinary Share",
                "CUSIP": "",
            },
            "KBANK.BK": {
                "BusinessEntity": "INSTRUMENTxEQUITY",
                "DocumentTitle": "Kasikornbank PCL, Ordinary Share",
                "SEDOL": "6888783",
            },
        }
    }
    df = discovery_lookup_build_df(data)
    for row in df.itertuples():
        if row.DocumentTitle[0] == "MSFT.O":
            assert row.SEDOL is pd.NA
            assert isinstance(row.CUSIP, str)
            assert row.CUSIP == "594918104"
        if row.DocumentTitle[0] == "GOOG.O":
            assert row.SEDOL is pd.NA
            assert isinstance(row.CUSIP, str)
            assert row.CUSIP == ""
        if row.DocumentTitle[0] == "KBANK.BK":
            assert row.CUSIP is pd.NA
            assert isinstance(row.SEDOL, str)
            assert row.SEDOL == "6888783"


def test_definition_parameters_is_not_none():
    # given
    request_factory = SearchRequestFactory()

    # when
    body_parameters = request_factory.get_body_parameters(**{"top": 0})

    # then
    assert body_parameters["Top"] == 0



from datetime import datetime, timedelta

import allure
import pytest

from refinitiv.data.content.ipa.dates_and_calendars import count_periods
from refinitiv.data.content.ipa.dates_and_calendars.count_periods._count_periods_data_provider import (
    Period,
)
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.helpers import (
    check_response_status,
    get_async_response_from_definitions,
    check_non_empty_response_data,
    check_response_dataframe_contains_columns_names,
    check_extended_params_were_sent_in_request,
)

EXPECTED_COLUMN_NAMES = ["count", "tag", "tenor"]


first_definition = count_periods.Definition(
    tag="my request 1",
    start_date=timedelta(-180),
    end_date=datetime.now(),
    period_type=count_periods.PeriodType.WORKING_DAY,
    calendars=["BAR", "KOR", "JAP"],
    currencies=["USD"],
    day_count_basis="Dcb_30E_360_ISMA",
)

second_definition = count_periods.Definition(
    tag="my request 2",
    start_date="2020-04-24",
    end_date="2021-04-24",
    period_type="NonWorkingDay",
    calendars=["BAR", "KOR", "JAP"],
    currencies=["USD"],
    day_count_basis=count_periods.DayCountBasis.DCB_30_ACTUAL,
)

invalid_definition = count_periods.Definition(
    tag="my request 2",
    start_date="2023.04.24",
    end_date=timedelta(0),
    period_type="Day",
)


@allure.suite("Dates And Calendars")
@allure.feature("Content object - Count Periods")
@allure.severity(allure.severity_level.CRITICAL)
class TestCountPeriods:
    @allure.title("Create a count_periods definition with valid params")
    @pytest.mark.caseid("C45567339")
    @pytest.mark.parametrize(
        "start_date, end_date, calendars, currencies, day_count_basis, period_type",
        [
            (
                "2018-12-31",
                "2020-12-31",
                ["UKR", "FRA"],
                ["EUR"],
                count_periods.DayCountBasis.DCB_30_360_US,
                count_periods.PeriodType.DAY,
            ),
            (
                timedelta(-180),
                timedelta(0),
                ["UKR", "FRA"],
                None,
                "Dcb_30_365_Brazil",
                "Year",
            ),
            (
                datetime(2020, 4, 13),
                datetime.now(),
                None,
                ["USD"],
                None,
                None,
            ),
        ],
    )
    def test_add_periods_definition_with_valid_params(
        self,
        open_session,
        start_date,
        end_date,
        calendars,
        currencies,
        day_count_basis,
        period_type,
    ):
        response = count_periods.Definition(
            tag="my request",
            period_type=period_type,
            start_date=start_date,
            end_date=end_date,
            calendars=calendars,
            currencies=currencies,
            day_count_basis=day_count_basis,
        ).get_data()

        assert isinstance(response.data.counted_period, Period)
        assert response.data.counted_period.tag == "my request"
        assert isinstance(response.data.counted_period.count, float)

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_response_dataframe_contains_columns_names(response, EXPECTED_COLUMN_NAMES)

    @allure.title(
        "Create a count_periods definition with valid and invalid definition asynchronously"
    )
    @pytest.mark.caseid("C46606039")
    async def test_count_periods_definition_with_valid_and_invalid_def_async(
        self, open_session_async
    ):
        valid_response, invalid_response = await get_async_response_from_definitions(
            first_definition, invalid_definition
        )

        check_non_empty_response_data(valid_response)
        check_response_status(
            response=valid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
        )

        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.FOUR_HUNDRED,
            expected_http_reason=HttpReason.BAD_REQUEST,
            expected_error_code=HttpStatusCode.FOUR_HUNDRED,
            expected_error_message="Business Calendar: end date must be greater than start date.",
        )

    @allure.title(
        "Create a count_periods definitions with valid and invalid definition and getting RDError"
    )
    @pytest.mark.parametrize(
        "expected_error",
        [
            (
                "Error code 400 | Business Calendar: end date must be greater than start date."
            )
        ],
    )
    @pytest.mark.caseid("C45567347")
    def test_count_periods_definitions_with_valid_and_invalid_def(
        self, open_session, expected_error
    ):
        with pytest.raises(RDError) as error:
            count_periods.Definitions(
                [first_definition, second_definition, invalid_definition]
            ).get_data()
        assert str(error.value) == expected_error

    @allure.title(
        "Create a count_periods definitions with valid and invalid definition params asynchronously"
    )
    @pytest.mark.caseid("C46606806")
    async def test_count_periods_definitions_with_valid_and_invalid_def_async(
        self, open_session_async
    ):
        valid_response, invalid_response = await get_async_response_from_definitions(
            count_periods.Definitions([first_definition, second_definition]),
            count_periods.Definitions(invalid_definition),
        )

        check_non_empty_response_data(valid_response)
        check_response_status(
            response=valid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
        )

        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.FOUR_HUNDRED,
            expected_http_reason=HttpReason.BAD_REQUEST,
            expected_error_code=HttpStatusCode.FOUR_HUNDRED,
            expected_error_message="Business Calendar: end date must be greater than start date.",
        )

    @allure.title("Create count_periods definition with extended params and get data")
    @pytest.mark.parametrize(
        "extended_param",
        [
            {
                "tag": "new tag",
                "startDate": "2022-07-04",
            }
        ],
    )
    @pytest.mark.caseid("C45567341")
    def test_count_periods_definition_with_extended_params(
        self, open_session, extended_param
    ):
        response = count_periods.Definition(
            tag="my request",
            period_type=count_periods.PeriodType.NON_WORKING_DAY,
            start_date=timedelta(-360),
            end_date=datetime.now(),
            calendars=["UKR", "FRA"],
            currencies=["EUR"],
            extended_params=extended_param,
        ).get_data()

        check_extended_params_were_sent_in_request(response, extended_param)
        check_non_empty_response_data(response)
        check_response_status(
            response=response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
        )

    @allure.title("Create an count_periods definition without session and get error")
    @pytest.mark.parametrize(
        "expected_error",
        ["No default session created yet. Please create a session first!"],
    )
    @pytest.mark.caseid("C45567348")
    def test_count_periods_definition_without_session(self, expected_error):
        with pytest.raises(AttributeError) as error:
            first_definition.get_data()
        assert str(error.value) == expected_error



from astroid import nodes
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker


class PrintChecker(BaseChecker):
    __implements__ = IAstroidChecker

    name = "print-checker"
    priority = -1
    msgs = {
        "W0001": (
            "'print' statement in project code.",
            "no-print-statement",
            "Do not use print statement in project code.",
        )
    }
    options = ()

    def visit_call(self, node: nodes.Call) -> None:
        if isinstance(node.func, nodes.Name) and node.func.name == "print":
            self.add_message("no-print-statement", node=node)


def register(linter):
    print("Registering PrintChecker...")
    checker = PrintChecker(linter)
    linter.register_checker(checker)
    print("...PrintChecker registered")
